import numpy as np
import healpy as hp

def recombine(groups, sliced_array):
    combined = np.array([])
    for group in groups:
        for i in range(len(sliced_array)):
            if i in group:
                combined

def similar_range(params, map_struct):

    if params['doObservability']:
        observability_struct = map_struct['observability']
        telescope = observability_struct.keys()[0]
        prob = observability_struct[telescope]['prob']

    else:
        prob = map_struct['prob']
    
    nested_map = hp.pixelfunc.reorder(prob, r2n = True)
    regions = params['Nregions']
    region_nsides = int((regions / 12) ** .5)
    res = hp.get_map_size(prob)
    region_size = res / regions
    sliced_array = np.zeros([regions, region_size])

    start = 0
    end = start + region_size

    for region in range(regions):
        sliced_array[region] = nested_map[start:end]
        start += region_size
        end += region_size

    # Theres def a way to do this with numpy but i'm on a train, can't see docs
    sum_hp = lambda arr: np.array([np.sum(row) for row in arr])

    sums = sum_hp(sliced_array)
    sum_neighbours = lambda n: np.sum([sums[p] for p in hp.get_all_neighbours(region_nsides, n, nest=True)])
    nb_sums = np.array([sums[i] + sum_neighbours(i) for i in range(len(sums))])
    region_order = np.argsort(nb_sums)[::-1]
    
    significant_regions = np.array([])
    for idx in region_order:
        if sums[idx] > max(sums) * 0.0001:
            significant_regions = np.append(significant_regions, idx)
    significant_regions = significant_regions.astype(int)

    groups = []

    def group_neighbours(group):
        neighbours = np.array([])
        for i in group:
            neighbours = np.append(neighbours, hp.get_all_neighbours(region_nsides, int(i), nest = True))
        return neighbours

    for idx in significant_regions:
        if len(groups) == 0:
            groups.append([idx])
            continue
        for group in groups:
            if idx in group_neighbours(group):
                group.append(idx)
                break
        else:
            groups.append([idx])
    
    group_maps = []
    for group in groups:
        combined_map = np.array([])
        for i in range(len(sliced_array)):
            if i in group:
                combined_map = np.append(combined_map, sliced_array[i])
            else: combined_map = np.append(combined_map, np.zeros(len(sliced_array[0])))
        group_maps.append(combined_map)

    group_maps = [hp.reorder(hp_map, n2r = True) for hp_map in group_maps]

    return group_maps


# Copyright (C) 2019 Siddharth Mohite
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from __future__ import print_function
import numpy as np
import healpy as hp
from astropy.table import Table
from astropy import constants as c, units as u
from astropy.coordinates import SkyCoord, ICRS, FK5
from astropy.io import fits
from astropy.wcs import WCS
import astropy
import healpy
# import cPickle as pickle
import pickle
import os

class DECamtile:
    def __init__(self, RA, Dec, number=None, missing=None):
        """
        z=DECamtile(RA, Dec, number, missing)
        creates a DECamtile object centered at the given coordinates
        coordinates can be Quantities (with units) or floats (assumed degrees)
        number is just used for informational purposes (bookkeeping)
        missing lists the missing ccds (potentially None or [])
        if missing is not None, should be a list of ccds (1-62)
        they will be removed from the ccd-by-ccd WCS

        """
        self.ccd_size=np.array([2046,4094])
#        self.quadrant_scale=0.27*u.arcsec

        if isinstance(RA, astropy.units.quantity.Quantity):            
            self.RA=RA
        else:
            self.RA=RA*u.deg
        if isinstance(Dec, astropy.units.quantity.Quantity):            
            self.Dec=Dec
        else:
            self.Dec=Dec*u.deg

        self.number=number

        centers = self.ccd_centers()
        self.ccd_RA=np.array([centers[key][0] for key in centers.keys()])*u.deg
        self.ccd_Dec=np.array([centers[key][1] for key in centers.keys()])*u.deg
        cd = np.array([[-1.60128078e-07,  7.28644399e-05],
                      [-7.28804334e-05, -1.41273307e-07]]) 
        
        self.missing_ccds=[]
        if missing is not None and len(missing)>0:
            for m in missing:
                if isinstance(m,int) or len(m)==1:
                    # assume this is 0 origin
                    self.missing_ccds.append(m)
#                elif len(m)==2:
                    # this assumes CCD and quadrant numbers are 1-origin
#                    self.missing_quadrants.append(4*(m[0]-1)+(m[1]-1))
        good_ccds=np.setdiff1d(np.arange(len(self.ccd_RA)),
                                    np.array(self.missing_ccds))
        
        self._wcs=[]
        for i in good_ccds:
            self._wcs.append(self.ccd_WCS(cd,self.ccd_RA[i],self.ccd_Dec[i]))

    def ccd_WCS(self, cd, ccd_RA, ccd_Dec):
        """
        w=ccd_WCS(cd,crpix,ccd_RA, ccd_Dec)
        returns a WCS object that is specific to the ccd RA and Dec specified
        overall scale, size are determined by class variables
        """
        
        w=WCS(naxis=2)
        w.wcs.crpix=(self.ccd_size+1.0)/2
        w.wcs.ctype=['RA---TAN','DEC--TAN']
        w.wcs.crval=[ccd_RA.value,
                     max(min(ccd_Dec.value,90),-90)]
        w.wcs.cd=cd
        return w

    def ccd_centers(self):
        """
        alpha,delta=ccd_centers()
        return celestial coordinates for the centers of each ccd
        given the pointing center RA,Dec
        """
        ccd_centers_radec = ccd_xy_to_radec(self.RA,self.Dec,self.get_ccd_centers())
        return ccd_centers_radec

    def get_ccd_centers(self):
        ccd_cent_xy = {'1': [-0.30848869614381674, 0.9048868132934124],'3': [0.3126638732505829, 0.903736681246994],
                       '4': [-0.46424313388432176, 0.7410089796458399],'5': [-0.15364984859501787, 0.7408713142103732],
                       '6': [0.1571387259082878, 0.7402269739588593],'7': [0.467603613257108, 0.7392171229774838],
                       '8': [-0.6198946868359682, 0.5768134685376444],'9': [-0.30955498335035025, 0.5768085523046568],
                       '10': [0.0013477793249647779, 0.5763113502103236],'11': [0.3121902019689574, 0.5755750278008702],
                       '12': [0.6226930307061053, 0.5743855934179499],'13': [-0.7755501585170581, 0.4124544450213818],
                       '14': [-0.46558119678543103, 0.41251128825254235],'15': [-0.15464112242830574, 0.41229197585181315],
                       '16': [0.15656398568088345, 0.4114606530655467],'17': [0.4674414894955094, 0.41072825795192136],
                       '18': [0.7777331864109138, 0.409567932442902],'19': [-0.7759834336369162, 0.24831566877921676],
                       '20': [-0.46607502185988525, 0.24782905120552876],'21': [-0.15509932243638994, 0.24758285798181506],
                       '22': [0.15626265250880025, 0.24702609588409785],'23': [0.4674434728526691, 0.2461625655342552],
                       '24': [0.777482221168825, 0.24521534351199428],'25': [-0.9314559066649502, 0.08383480908831277],
                       '26': [-0.6216994874177798, 0.08353996510188882],'27': [-0.31102353489225215, 0.08283960916939065],
                       '28': [0.0005071693155911257, 0.08241792995217975],'29': [0.31158305482235177, 0.08163431022316624],
                       '30': [0.6224533719596026, 0.08098146395460736],'31': [0.9320777373043198, 0.08044138862297144],
                       '32': [-0.9319464165707169, -0.0805589216843629],'33': [-0.622133652371652, -0.08122227572955273],
                       '34': [-0.31148773321326756, -0.08193206187722929],'35': [-0.00014499352178874166, -0.08254844513458418],
                       '36': [0.31098135984096015, -0.08317938096258874],'37': [0.6217138680647583, -0.08356797238031384],
                       '38': [0.9315841692181391, -0.08418488728808847],'39': [-0.7773609424381199, -0.245412819661229],
                       '40': [-0.46728294158166866, -0.24630563002879535],'41': [-0.15615176138477885, -0.2469488238588537],
                       '42': [0.15523767984355089, -0.24768250766829886],'43': [0.4662187887667616, -0.24809293832066254],
                       '44': [0.776246372428461, -0.24844509764265615],'45': [-0.7775470321940903, -0.40967055522534995],
                       '46': [-0.4676908153361856, -0.41077195744348843],'47': [-0.15653339105153957, -0.41170581376861437],
                       '48': [0.15449971354932118, -0.41233044750284964],'49': [0.46565358463600437, -0.41260284266708563],
                       '50': [0.7757206230994781, -0.41269011164669245],'51': [-0.6225729237947966, -0.5746346822329046],
                       '52': [-0.3122155321260558, -0.5755970743456301],'53': [-0.0011401216990000642, -0.5763197415054745],
                       '54': [0.30968238386568164, -0.5769630649576271],'55': [0.6200804504434518, -0.5769886473852198],
                       '56': [-0.46754566383013146, -0.7392890744530034],'57': [-0.15717538834988146, -0.7402009474290661],
                       '58': [0.15377123111958424, -0.7408432843903323],'59': [0.4642380873704591, -0.7410163636305219],
                       '60': [-0.3125067998457982, -0.9038265779307155],'62': [0.3086296599756487, -0.9050452408374547]}
        return ccd_cent_xy

class CCDProb:
    '''
    Class :: Instantiate a CCDProb object that will allow us to calculate the
             probability content in a single CCD.
    '''
    def __init__(self, RA, Dec):
        self.RA = RA
        self.Dec = Dec
        self.ccd_size = np.array([2046, 4094])

def get_decam_ccds(ra,dec,save_footprint=False):
    """Calculate DECam CCD footprints as offsets from the telescope
    boresight. Also optionally saves the footprints in a region file 
    called footprint.reg"""
#    ccd_prob = CCDProb(ra, dec)
    decam_tile = DECamtile(ra, dec)
    ccd_cents_ra = decam_tile.ccd_RA
    ccd_cents_dec = decam_tile.ccd_Dec
    offsets = np.asarray([decam_tile._wcs[ccd_id].calc_footprint(axes=decam_tile.ccd_size)
            for ccd_id in range(len(ccd_cents_ra))])
    if save_footprint:
        ra,dec = np.transpose(offsets, (2, 0, 1))
        with open('footprint.reg','a') as f:
            for i in range(len(ra)): 
                lines = ['# Region file format: DS9 version 4.0 \n', '# global color=green font="helvetica 12 bold select=1 highlite=1 edit=1 move=1 delete=1 include=1 fixed=0 source \n', 'ICRS \n', 'polygon('+str(ra[i,0]) +',' + str(dec[i,0]) + ',' + str(ra[i,1]) +',' + str(dec[i,1]) + ',' + str(ra[i,2]) +',' + str(dec[i,2]) + ',' + str(ra[i,3]) +',' + str(dec[i,3]) + ') # color=green, width=2 \n'] 
                f.writelines(lines)
            f.close()
    return np.transpose(offsets, (2, 0, 1))

def get_quadrant_ipix(nside, ra, dec):

    ccd_coords = get_decam_ccds(0,0)

    skyoffset_frames = SkyCoord(ra, dec, unit=u.deg).skyoffset_frame()
    ccd_coords_icrs = SkyCoord(
                    *np.tile(
                        ccd_coords[:, np.newaxis, ...],
                        (1, 1, 1)), unit=u.deg,
                    frame=skyoffset_frames[:, np.newaxis, np.newaxis]
                ).transform_to(ICRS)
    ccd_xyz = np.moveaxis(
        ccd_coords_icrs.cartesian.xyz.value, 0, -1)[0]

    ipixs = []
    for subfield_id, xyz in enumerate(ccd_xyz):
        ipix = hp.query_polygon(nside, xyz)
        ipixs.append(ipix.tolist())
    return ipixs

def get_ccdcenters_radec(filename):
    hdul = fits.open(filename)
    nccds = len(hdul) - 1
    ccd_centers = dict()
    for i in range(1,nccds+1):
        w = WCS(hdul[i].header)
        ccdnum = hdul[i].header['CCDNUM']
        coords = np.array(w.wcs_pix2world(hdul[i].header['NAXIS1']/2,hdul[i].header['NAXIS2']/2,0))
        ccd_centers[str(ccdnum)] = coords
    hdul.close()
    return ccd_centers

def create_region_file(filename):
    hdul = fits.open(filename)
    nccds = len(hdul) - 1
    for i in range(1,nccds+1):
        w = WCS(hdul[i].header)
        w.footprint_to_file(filename='region_files/footprint_'+str(hdul[i].header['CCDNUM'])+'.reg')
    hdul.close()

def get_cd(filename):
    hdul = fits.open(filename)
    w = WCS(hdul[1].header)
    hdul.close()
    return w.wcs.cd

def get_ccdcenters_xy(filename):
    "Calculates the x,y coordinates of the CCD center given the RA,Dec of the center and reference point.Uses Calabretta and Greisen(2002) eqns. 5,12,13,54"
    ccd_centers_xy = dict()
    phi_p = 180*u.deg
    ccd_centers_radec = get_ccdcenters_radec(filename)
    hdul = fits.open(filename)
    alpha_p = hdul[1].header['CRVAL1']*u.deg
    delta_p = hdul[1].header['CRVAL2']*u.deg
    ccdnums = ccd_centers_radec.keys()
    for ccdnum in ccdnums:
        alpha = ccd_centers_radec[ccdnum][0]*u.deg
        delta = ccd_centers_radec[ccdnum][1]*u.deg
        phi = phi_p + np.arctan2(-np.cos(delta)*np.sin(alpha-alpha_p),np.sin(delta)*np.cos(delta_p)-np.cos(delta)*np.sin(delta_p)*np.cos(alpha-alpha_p))
        theta = np.arcsin(np.sin(delta)*np.sin(delta_p) + np.cos(delta)*np.cos(delta_p)*np.cos(alpha-alpha_p))
        R_theta = (1*u.rad/np.tan(theta)).to(u.deg)
        x = R_theta * np.sin(phi)
        y = -R_theta * np.cos(phi)
        ccd_centers_xy[ccdnum] = [x.value,y.value]
    return ccd_centers_xy

def ccd_xy_to_radec(alpha_p,delta_p,ccd_centers_xy):
    # convert to Native longitude (phi) and latitude (theta)
    # need intermediate step (Rtheta) to deal with TAN projection
    # Calabretta & Greisen (2002), Eqn. 14,15
    alpha_p = alpha_p
    delta_p = delta_p
    ccd_centers_radec = dict()
    for ccdnum in ccd_centers_xy.keys():
        x = ccd_centers_xy[ccdnum][0]*u.deg
        y = ccd_centers_xy[ccdnum][1]*u.deg
        phi=np.arctan2(x,-y)
        Rtheta=np.sqrt(x**2+y**2)
        # now to theta using the TAN projection
        # Calabrett & Greisen (2002), Eqn. 55
        theta=np.arctan2(1,Rtheta.to(u.rad).value)*u.rad

        # Native longitude/latitude of fiducial point
        # appropriate for zenithal projections including TAN
        phi0=0*u.deg
        theta0=90*u.deg
        # Native longitude/latitue of celestial pole
        # for delta0<theta0 then phip should be 180 deg
        # and theta0=90 deg
        phi_p=180*u.deg
        theta_p=delta_p

        # Celestial longitude/latitude
        # Calabretta & Greisen (2002), Eqn. 2
        alpha=alpha_p+np.arctan2(-np.cos(theta)*np.sin(phi-phi_p),
                                 np.sin(theta)*np.cos(delta_p)-np.cos(theta)*np.sin(delta_p)*np.cos(phi-phi_p))
        delta=(np.arcsin(np.sin(theta)*np.sin(delta_p)+np.cos(theta)*np.cos(delta_p)*np.cos(phi-phi_p))).to(u.deg)
        alpha[alpha<0*u.deg]+=360*u.deg
        ccd_centers_radec[ccdnum] = np.array([alpha.value,delta.value])
    return ccd_centers_radec

def calculate_residuals(ccd_cent_code,ccd_cent_wcs):
    residuals = np.empty([len(ccd_cent_code),2])
    for i,ccdnum in enumerate(ccd_cent_code.keys()):
        residuals[i] = np.array([np.abs(ccd_cent_code[ccdnum][0]-ccd_cent_wcs[ccdnum][0]),np.abs(ccd_cent_code[ccdnum][1]-ccd_cent_wcs[ccdnum][1])])
    return residuals





import os, sys, copy
import numpy as np
import healpy as hp

from VOEventLib.VOEvent import Table, Field, What
from VOEventLib.Vutil import utilityTable, stringVOEvent, VOEventExportClass

import astropy.coordinates
from astropy.time import Time, TimeDelta
import astropy.units as u
import ligo.segments as segments
import gwemopt.utils
import gwemopt.rankedTilesGenerator
import gwemopt.coverage
from gwemopt.segments import angular_distance
from munkres import Munkres, make_cost_matrix

def get_altaz_tiles(ras, decs, observatory, obstime):

    # Convert to RA, Dec.
    radecs = astropy.coordinates.SkyCoord(
            ra=np.array(ras)*u.degree, dec=np.array(decs)*u.degree, frame='icrs')

    # Alt/az reference frame at observatory, now
    frame = astropy.coordinates.AltAz(obstime=obstime, location=observatory)
            
    # Transform grid to alt/az coordinates at observatory, now
    altaz = radecs.transform_to(frame)    

    return altaz


def find_tile(exposureids_tile, exposureids, probs, idxs=None,
              exptimecheckkeys=[], current_ra=np.nan, current_dec=np.nan,
              slew_rate=1, readout=1):
    # exposureids_tile: {expo id}-> list of the tiles available for observation
    # exposureids: list of tile ids for every exposure it is allocated to observe
    if idxs is not None:
        for idx in idxs:
            if len(exposureids_tile["exposureids"])-1 < idx: continue
            idx2 = exposureids_tile["exposureids"][idx]
            if idx2 in exposureids and not idx2 in exptimecheckkeys:
                idx = exposureids.index(idx2)
                exposureids.pop(idx)
                probs.pop(idx)
                return idx2, exposureids, probs

    findTile = True
    while findTile:
        if not exposureids_tile["probs"]:
            idx2 = -1
            findTile = False
            break
        if (not np.isnan(current_ra)) and (not np.isnan(current_dec)):
            dist = angular_distance(current_ra, current_dec,
                                    np.array(exposureids_tile["ras"]),
                                    np.array(exposureids_tile["decs"]))
            slew_readout = readout/(dist/slew_rate)
            slew_readout[slew_readout>1] = 1.0
            score = np.array(exposureids_tile["probs"]) * slew_readout 
            idx = np.argmax(score)
        else:
            idx = np.argmax(exposureids_tile["probs"])
        idx2 = exposureids_tile["exposureids"][idx]

        if exposureids:
            if idx2 in exposureids and not idx2 in exptimecheckkeys:
                idx = exposureids.index(idx2)
                exposureids.pop(idx)
                probs.pop(idx)
                findTile = False
            else:
                exposureids_tile["exposureids"].pop(idx)
                exposureids_tile["probs"].pop(idx)
                exposureids_tile["ras"].pop(idx)
                exposureids_tile["decs"].pop(idx)
        else:
            findTile = False

    return idx2, exposureids, probs

def get_order(params, tile_struct, tilesegmentlists, exposurelist, observatory, config_struct):
    '''
    tile_struct: dictionary. key -> struct info. 
    tilesegmentlists: list of lists. Segments for each tile in tile_struct 
        that are available for observation.
    exposurelist: list of segments that the telescope is supposed to be working.
        consecutive segments from the start to the end, with each segment size
        being the exposure time.
    Returns a list of tile indices in the order of observation.
    '''
    keys = tile_struct.keys()

    exposureids_tiles = {}
    first_exposure = np.inf*np.ones((len(keys),))
    last_exposure = -np.inf*np.ones((len(keys),))
    tileprobs = np.zeros((len(keys),))
    tilenexps = np.zeros((len(keys),))
    tileexptime = np.zeros((len(keys),))
    tilefilts = {}
    tileavailable = np.zeros((len(keys),))
    tileavailable_tiles = {} 
    keynames = []

    nexps = 0
    for jj, key in enumerate(keys):
        tileprobs[jj] = tile_struct[key]["prob"]
        tilenexps[jj] = tile_struct[key]["nexposures"]
        tilefilts[key] = copy.deepcopy(tile_struct[key]["filt"])
        tileavailable_tiles[jj] = []
        keynames.append(key) 

        nexps = nexps + tile_struct[key]["nexposures"]

    if "dec_constraint" in config_struct:
        dec_constraint = config_struct["dec_constraint"].split(",")
        dec_min = float(dec_constraint[0])
        dec_max = float(dec_constraint[1])

    for ii in range(len(exposurelist)):
        exposureids_tiles[ii] = {}
        exposureids = []
        probs = []
        ras, decs = [], []
        for jj, key in enumerate(keys):
            tilesegmentlist = tilesegmentlists[jj]
            if tile_struct[key]["prob"] == 0: continue
            if "dec_constraint" in config_struct:
                if (tile_struct[key]["dec"] < dec_min) or (tile_struct[key]["dec"] > dec_max):
                    continue
            if "epochs" in tile_struct[key]:
                if np.any(np.abs(exposurelist[ii][0]-tile_struct[key]["epochs"][:,2]) < params["mindiff"]/86400.0):
                    continue
            if tilesegmentlist.intersects_segment(exposurelist[ii]):
                exposureids.append(key)
                probs.append(tile_struct[key]["prob"])
                ras.append(tile_struct[key]["ra"])
                decs.append(tile_struct[key]["dec"])

                first_exposure[jj] = np.min([first_exposure[jj],ii])
                last_exposure[jj] = np.max([last_exposure[jj],ii])
                tileavailable_tiles[jj].append(ii)
                tileavailable[jj] = tileavailable[jj] + 1
        # in every exposure, the tiles available for observation
        exposureids_tiles[ii]["exposureids"] = exposureids # list of tile ids
        exposureids_tiles[ii]["probs"] = probs # the corresponding probs
        exposureids_tiles[ii]["ras"] = ras
        exposureids_tiles[ii]["decs"] = decs

    exposureids = []
    probs = []
    ras, decs = [], []
    for ii, key in enumerate(keys):
        # tile_struct[key]["nexposures"]: the number of exposures assigned to this tile
        for jj in range(tile_struct[key]["nexposures"]): 
            exposureids.append(key) # list of tile ids for every exposure it is allocated to observe
            probs.append(tile_struct[key]["prob"])
            ras.append(tile_struct[key]["ra"])
            decs.append(tile_struct[key]["dec"])

    idxs = -1*np.ones((len(exposureids_tiles.keys()),))
    filts = ['n'] * len(exposureids_tiles.keys())

    if nexps == 0:
        return idxs, filts    

    if params["scheduleType"] == "airmass_weighted":

        # # first step is to sort the array in order of descending probability
        indsort = np.argsort(-np.array(probs))
        probs = np.array(probs)[indsort]
        ras = np.array(ras)[indsort]
        decs = np.array(decs)[indsort]
        exposureids = np.array(exposureids)[indsort] 

        tilematrix = np.zeros((len(exposurelist), len(ras)))
        probmatrix = np.zeros((len(exposurelist), len(ras)))

        for ii in np.arange(len(exposurelist)):

            # first, create an array of airmass-weighted probabilities
            t = Time(exposurelist[ii][0], format='mjd')
            altaz = get_altaz_tiles(ras, decs, observatory, t)
            alts = altaz.alt.degree
            horizon = config_struct["horizon"]
            horizon_mask = alts <= horizon
            airmass = 1 / np.cos((90. - alts) * np.pi / 180.)
            below_horizon_mask = horizon_mask * 10.**100
            airmass = airmass + below_horizon_mask
            airmass_weight = 10 ** (0.4 * 0.1 * (airmass - 1) )
            tilematrix[ii, :] = np.array(probs/airmass_weight)
            probmatrix[ii, :] = np.array(probs * (True^horizon_mask))

    if params["scheduleType"] == "greedy":
        for ii in np.arange(len(exposurelist)): 
            exptimecheck = np.where(exposurelist[ii][0]-tileexptime <
                                    params["mindiff"]/86400.0)[0]
            exptimecheckkeys = [keynames[x] for x in exptimecheck]

            # find_tile finds the tile that covers the largest probablity
            # restricted by availability of tile and timeallocation
            idx2, exposureids, probs = find_tile(exposureids_tiles[ii],exposureids,probs,exptimecheckkeys=exptimecheckkeys)
            if idx2 in keynames:
                idx = keynames.index(idx2)
                tilenexps[idx] = tilenexps[idx] - 1
                tileexptime[idx] = exposurelist[ii][0]
                if len(tilefilts[idx2]) > 0:
                    filt = tilefilts[idx2].pop(0)
                    filts[ii] = filt
            idxs[ii] = idx2

            if not exposureids: break
    elif params["scheduleType"] == "greedy_slew":
        current_ra, current_dec = np.nan, np.nan
        for ii in np.arange(len(exposurelist)):
            exptimecheck = np.where(exposurelist[ii][0]-tileexptime <
                                    params["mindiff"]/86400.0)[0]
            exptimecheckkeys = [keynames[x] for x in exptimecheck]

            # find_tile finds the tile that covers the largest probablity
            # restricted by availability of tile and timeallocation
            idx2, exposureids, probs = find_tile(exposureids_tiles[ii],exposureids,probs,exptimecheckkeys=exptimecheckkeys,current_ra=current_ra,current_dec=current_dec,slew_rate=config_struct['slew_rate'],readout=config_struct['readout'])
            if idx2 in keynames:
                idx = keynames.index(idx2)
                tilenexps[idx] = tilenexps[idx] - 1
                tileexptime[idx] = exposurelist[ii][0]
                if len(tilefilts[idx2]) > 0:
                    filt = tilefilts[idx2].pop(0)
                    filts[ii] = filt
                current_ra = tile_struct[idx2]["ra"]
                current_dec = tile_struct[idx2]["dec"]

            idxs[ii] = idx2

            if not exposureids: break
    elif params["scheduleType"] == "sear":
        #for ii in np.arange(len(exposurelist)):
        iis = np.arange(len(exposurelist)).tolist()
        while len(iis) > 0: 
            ii = iis[0]
            mask = np.where((ii == last_exposure) & (tilenexps > 0))[0]

            exptimecheck = np.where(exposurelist[ii][0]-tileexptime <
                                    params["mindiff"]/86400.0)[0]
            exptimecheckkeys = [keynames[x] for x in exptimecheck]

            if len(mask) > 0:
                idxsort = mask[np.argsort(tileprobs[mask])]
                idx2, exposureids, probs = find_tile(exposureids_tiles[ii],exposureids,probs,idxs=idxsort,exptimecheckkeys=exptimecheckkeys) 
                last_exposure[mask] = last_exposure[mask] + 1
            else:
                idx2, exposureids, probs = find_tile(exposureids_tiles[ii],exposureids,probs,exptimecheckkeys=exptimecheckkeys)
            if idx2 in keynames:
                idx = keynames.index(idx2)
                tilenexps[idx] = tilenexps[idx] - 1
                tileexptime[idx] = exposurelist[ii][0]
                if len(tilefilts[idx2]) > 0:
                    filt = tilefilts[idx2].pop(0)
                    filts[ii] = filt
            idxs[ii] = idx2 
            iis.pop(0)

            if not exposureids: break

    elif params["scheduleType"] == "weighted":
        for ii in np.arange(len(exposurelist)):
            jj = exposureids_tiles[ii]["exposureids"]
            weights = tileprobs[jj] * tilenexps[jj] / tileavailable[jj]
            weights[~np.isfinite(weights)] = 0.0

            exptimecheck = np.where(exposurelist[ii][0]-tileexptime <
                                    params["mindiff"]/86400.0)[0]
            weights[exptimecheck] = 0.0

            if np.any(weights >= 0):
                idxmax = np.argmax(weights)
                idx2 = jj[idxmax]
                if idx2 in keynames:
                    idx = keynames.index(idx2)
                    tilenexps[idx] = tilenexps[idx] - 1
                    tileexptime[idx] = exposurelist[ii][0]
                    if len(tilefilts[idx2]) > 0:
                        filt = tilefilts[idx2].pop(0)
                        filts[ii] = filt
                idxs[ii] = idx2
            tileavailable[jj] = tileavailable[jj] - 1  

    elif params["scheduleType"] == "airmass_weighted":
        # then use the Hungarian algorithm (munkres) to schedule high prob tiles at low airmass
        tilematrix_mask = tilematrix > 10**(-10)

        if tilematrix_mask.any():
            print("Calculating Hungarian solution...")
            total_cost = 0
            cost_matrix = make_cost_matrix(tilematrix)
            m = Munkres()
            optimal_points = m.compute(cost_matrix)
            print("Hungarian solution calculated...")
            max_no_observ = min(tilematrix.shape)
            for jj in range(max_no_observ):
                idx0, idx1 = optimal_points[jj]
                # idx0 indexes over the time windows, idx1 indexes over the probabilities
                # idx2 gets the exposure id of the tile, used to assign tileexptime and tilenexps
                try:
                    idx2 = exposureids[idx1]
                    pamw = tilematrix[idx0][idx1]
                    total_cost += pamw
                    if len(tilefilts[idx2]) > 0:
                        filt = tilefilts[idx2].pop(0)
                        filts[idx0] = filt
                except: continue

                idxs[idx0] = idx2

        else: print("The localization is not visible from the site.")

    else:
        raise ValueError("Scheduling options are greedy/sear/weighted/airmass_weighted, or with _slew.")

    return idxs, filts


def get_order_slew(params, tile_struct, tilesegmentlists, config_struct):   
    keys = tile_struct.keys() 
    keynames = []
    namekeys = {}
    tilefilts = {}
    tileAllocatedTime = {}
    tileprobs = np.zeros((len(keys),))
    for jj, key in enumerate(keys):
        tileprobs[jj] = tile_struct[key]["prob"]
        tilefilts[key] = copy.deepcopy(tile_struct[key]["filt"])
        keynames.append(key) 
        namekeys[key] = jj
        tileAllocatedTime[key] = tile_struct[key]["exposureTime"] / 86400
    lastObs = np.zeros((len(keys),))

    idxs = []
    filts = []
    exposurelist = []

    if params["scheduleType"] == "greedy_slew":
        current_time = Time(params['gpstime'], format='gps', scale='utc').mjd + np.min(params["Tobs"][::2])
        current_ra = config_struct["latitude"]
        current_dec = config_struct["longitude"]
        idle = 0
        while True:
            # check the time gap since last observation
            exptimecheck = np.where(current_time - lastObs < params["mindiff"]/86400.0)[0]
            exptimecheckkeys = [keynames[x] for x in exptimecheck]
            idx2, slew_readout, exp_idle_seg = find_tile_greedy_slew(current_time, current_ra, current_dec, tilesegmentlists, tileprobs, 
                    config_struct, tile_struct, keynames, tileAllocatedTime, exptimecheckkeys=exptimecheckkeys, idle=idle)
            if idx2 != -1:
                exp_idle_len = exp_idle_seg[1] - exp_idle_seg[0]
                idle = 0
                idxs.append(idx2)
                exposurelist.append(exp_idle_seg)
                filts.append(tilefilts[idx2].pop(0))
                tileAllocatedTime[idx2] = tileAllocatedTime[idx2] - exp_idle_len
                current_time = exp_idle_seg[1]
                current_ra = tile_struct[idx2]["ra"]
                current_dec = tile_struct[idx2]["dec"]
                idx = namekeys[idx2]
                lastObs[idx] = current_time
            elif exp_idle_seg is not None:
                exp_idle_len = exp_idle_seg[1] - exp_idle_seg[0]
                idxs.append(idx2)
                exposurelist.append(exp_idle_seg)
                filts.append('n')
                idle = exp_idle_len
                current_time += exp_idle_len
            else:
                break
    elif params["scheduleType"] == "sear_slew":
        print('sear_slew is not ready yet.')
    elif params["scheduleType"] == "weighted_slew":
        print('weighted_slew is not ready yet.')
    else:
        raise ValueError("Scheduling options are greedy/sear/weighted, or with _slew.")

    return idxs, exposurelist, filts


def scheduler(params, config_struct, tile_struct):
    '''
    config_struct: the telescope configurations
    tile_struct: the tiles, contains time allocation information
    '''
    import gwemopt.segments
    #import gwemopt.segments_astroplan
    coverage_struct = {}
    coverage_struct["data"] = np.empty((0,8))
    coverage_struct["filters"] = []
    coverage_struct["ipix"] = []
    coverage_struct["patch"] = []
    coverage_struct["area"] = []
    if params["tilesType"] == "galaxy":
        coverage_struct["galaxies"] = []

    observatory = astropy.coordinates.EarthLocation(
        lat=config_struct["latitude"]*u.deg, lon=config_struct["longitude"]*u.deg, height=config_struct["elevation"]*u.m)

    segmentlist = config_struct["segmentlist"]
    exposurelist = config_struct["exposurelist"]
    #tilesegmentlists = gwemopt.segments_astroplan.get_segments_tiles(config_struct, tile_struct, observatory, segmentlist)
    tilesegmentlists = []
    keys = tile_struct.keys()
    for key in keys:
        # segments.py: tile_struct[key]["segmentlist"] is a list of segments when the tile is available for observation
        tilesegmentlists.append(tile_struct[key]["segmentlist"]) 
    print("Generating schedule order...")
    keys, filts = get_order(params,tile_struct,tilesegmentlists,exposurelist,observatory, config_struct)
    if params["doPlots"]:
        gwemopt.plotting.scheduler(params,exposurelist,keys)

    exposureused = np.where(np.array(keys)>=0)[0]
    coverage_struct["exposureused"] = exposureused
    while len(exposurelist) > 0:
        key, filt = keys[0], filts[0]
        if key == -1:
            keys = keys[1:]
            filts = filts[1:]
            exposurelist = exposurelist[1:]
        else:
            tile_struct_hold = tile_struct[key]
            mjd_exposure_start = exposurelist[0][0]
            nkeys = len(keys)
            for jj in range(nkeys):
                if (keys[jj] == key) and (filts[jj] == filt) and not (nkeys == jj+1):
                    mjd_exposure_end = exposurelist[jj][1]
                elif (keys[jj] == key) and (filts[jj] == filt) and (nkeys == jj+1):
                    mjd_exposure_end = exposurelist[jj][1]
                    nexp = jj + 1
                    keys = []
                    filts = []
                    exposurelist = []
                else:
                    nexp = jj + 1
                    keys = keys[jj:]
                    filts = filts[jj:]
                    exposurelist = exposurelist[jj:]
                    break    
 
            # the (middle) tile observation time is mjd_exposure_mid
            mjd_exposure_mid = (mjd_exposure_start+mjd_exposure_end)/2.0

            # calculate airmass for each tile at the start of its exposure:
            t = Time(mjd_exposure_start, format='mjd')
            altaz = get_altaz_tiles(tile_struct_hold["ra"], tile_struct_hold["dec"], observatory, t)
            alt = altaz.alt.degree
            airmass = 1 / np.cos((90. - alt) * np.pi / 180)

            # total duration of the observation (?)


            nmag = np.log(nexp) / np.log(2.5)
            mag = config_struct["magnitude"] + nmag
            exposureTime = (mjd_exposure_end-mjd_exposure_start)*86400.0

            coverage_struct["data"] = np.append(coverage_struct["data"],np.array([[tile_struct_hold["ra"],tile_struct_hold["dec"],mjd_exposure_start,mag,exposureTime,int(key),tile_struct_hold["prob"],airmass]]),axis=0)

            coverage_struct["filters"].append(filt)
            coverage_struct["patch"].append(tile_struct_hold["patch"])
            coverage_struct["ipix"].append(tile_struct_hold["ipix"])
            coverage_struct["area"].append(tile_struct_hold["area"])
            if params["tilesType"] == "galaxy":
                coverage_struct["galaxies"].append(tile_struct_hold["galaxies"])

    coverage_struct["area"] = np.array(coverage_struct["area"])
    coverage_struct["filters"] = np.array(coverage_struct["filters"])
    coverage_struct["FOV"] = config_struct["FOV"]*np.ones((len(coverage_struct["filters"]),))
    coverage_struct["telescope"] = [config_struct["telescope"]]*len(coverage_struct["filters"])

    return coverage_struct


def computeSlewReadoutTime(config_struct, coverage_struct):
    slew_rate = config_struct['slew_rate']
    readout = config_struct['readout']
    prev_ra = config_struct["latitude"]
    prev_dec = config_struct["longitude"]
    acc_time = 0
    for dat in coverage_struct['data']:
        dist = angular_distance(prev_ra, prev_dec,
                                dat[0], dat[1])
        slew_readout_time = np.max([dist/slew_rate, readout])
        acc_time += slew_readout_time
        prev_dec = dat[0]
        prev_ra = dat[1]
    return acc_time

def write_xml(xmlfile,map_struct,coverage_struct,config_struct):

    what = What()

    table = Table(name="data", Description=["The datas of GWAlert"])
    table.add_Field(Field(name=r"grid_id", ucd="", unit="", dataType="int", \
                    Description=["ID of the grid of fov"]))
    table.add_Field(Field(name="field_id", ucd="", unit="", dataType="int",\
                    Description=["ID of the filed"]))
    table.add_Field(
        Field(
            name=r"ra", ucd=r"pos.eq.ra ", unit="deg", dataType="float",
            Description=["The right ascension at center of fov in equatorial coordinates"]
            )
        )
    table.add_Field(
        Field(
            name="dec", ucd="pos.eq.dec ", unit="deg", dataType="float",
            Description=["The declination at center of fov in equatorial coordinates"]
            )
        )
    table.add_Field(
        Field(
            name="ra_width", ucd=" ", unit="deg", dataType="float",
            Description=["Width in RA of the fov"]
            )
        )
    table.add_Field(
        Field(
            name="dec_width", ucd="", unit="deg", dataType="float",
            Description=["Width in Dec of the fov"]
            )
        )
    table.add_Field(
        Field(
            name="prob_sum", ucd="", unit="None", dataType="float",
            Description=["The sum of all pixels in the fov"]
            )
        )
    table.add_Field(
        Field(
            name="observ_time", ucd="", unit="sec", dataType="float",
            Description=["Tile mid. observation time in MJD"]
            )
        )
    table.add_Field(
        Field(
            name="airmass", ucd="", unit="None", dataType="float",
            Description=["Airmass of tile at mid. observation time"]
            )
        )
    table.add_Field(Field(name="priority", ucd="", unit="", dataType="int", Description=[""]))
    table_field = utilityTable(table)
    table_field.blankTable(len(coverage_struct["ipix"]))

    for ii in range(len(coverage_struct["ipix"])):
        data = coverage_struct["data"][ii,:]
        filt = coverage_struct["filters"][ii]
        ipix = coverage_struct["ipix"][ii]
        patch = coverage_struct["patch"][ii]
        FOV = coverage_struct["FOV"][ii]
        area = coverage_struct["area"][ii]

        prob = np.sum(map_struct["prob"][ipix])

        ra, dec = data[0], data[1]
        observ_time, exposure_time, field_id, prob, airmass = data[2], data[4], data[5], data[6], data[7]

        table_field.setValue("grid_id", ii, 0)
        table_field.setValue("field_id", ii, field_id)
        table_field.setValue("ra", ii, ra)
        table_field.setValue("dec", ii, dec)
        table_field.setValue("ra_width", ii, config_struct["FOV"])
        table_field.setValue("dec_width", ii, config_struct["FOV"])
        table_field.setValue("observ_time", ii, observ_time)
        table_field.setValue("airmass", ii, airmass)
        table_field.setValue("prob_sum", ii, prob)
        table_field.setValue("priority", ii, ii)
    table = table_field.getTable()
    what.add_Table(table)
    xml = stringVOEvent(what)
    lines = xml.splitlines()
    linesrep = []
    for line in lines:
        linenew = line.replace(">b'",">").replace("'</","</").replace("=b'","=").replace("'>",">")
        linesrep.append(linenew)
    xmlnew = "\n".join(linesrep)
    fid = open(xmlfile, "w")
    fid.write(xmlnew)
    fid.close()

def summary(params, map_struct, coverage_struct):

    idx50 = len(map_struct["cumprob"])-np.argmin(np.abs(map_struct["cumprob"]-0.50))
    idx90 = len(map_struct["cumprob"])-np.argmin(np.abs(map_struct["cumprob"]-0.90))

    mapfile = os.path.join(params["outputDir"],'map.dat')
    fid = open(mapfile,'w')
    fid.write('%.5f %.5f\n'%(map_struct["pixarea_deg2"]*idx50,map_struct["pixarea_deg2"]*idx90))
    fid.close()
    filts = list(set(coverage_struct["filters"]))
    for jj, telescope in enumerate(params["telescopes"]):

        schedulefile = os.path.join(params["outputDir"],'schedule_%s.dat'%telescope)
        schedulexmlfile = os.path.join(params["outputDir"],'schedule_%s.xml'%telescope)
        config_struct = params["config"][telescope]

        write_xml(schedulexmlfile,map_struct,coverage_struct,config_struct)

        if (params["tilesType"] == "hierarchical") or (params["tilesType"] == "greedy"):
            fields = np.zeros((params["Ntiles"][jj],len(filts)+2))
        else:
            fields = np.zeros((len(config_struct["tesselation"]),len(filts)+2))

        fid = open(schedulefile,'w')
        for ii in range(len(coverage_struct["ipix"])):
            if not telescope == coverage_struct["telescope"][ii]:
                continue

            data = coverage_struct["data"][ii,:]
            filt = coverage_struct["filters"][ii]
            ipix = coverage_struct["ipix"][ii]
            patch = coverage_struct["patch"][ii]
            FOV = coverage_struct["FOV"][ii]
            area = coverage_struct["area"][ii]
            rand = np.random.randint(2)
            
            prob = np.sum(map_struct["prob"][ipix])

            ra, dec = data[0], data[1]
            observ_time, exposure_time, field_id, prob, airmass = data[2], data[4], data[5], data[6], data[7]
            fid.write('%d %.5f %.5f %.5f %d %.5f %.5f %s %d\n'%(field_id,ra,dec,observ_time,exposure_time,prob,airmass,filt,rand))

            dist = angular_distance(data[0], data[1],
                                    config_struct["tesselation"][:,1],
                                    config_struct["tesselation"][:,2])
            idx1 = np.argmin(dist)
            idx2 = filts.index(filt)
            fields[idx1,0] = config_struct["tesselation"][idx1,0]
            fields[idx1,1] = prob
            fields[idx1,idx2+2] = fields[idx1,idx2+2]+1
        fid.close()

        idx = np.where(fields[:,1]>0)[0]
        fields = fields[idx,:]
        idx = np.argsort(fields[:,1])[::-1]
        fields = fields[idx,:]

        fields_sum = np.sum(fields[:,2:],axis=1)
        idx = np.where(fields_sum >= 2)[0]
        print('%d/%d fields were observed at least twice\n'%(len(idx),len(fields_sum)))

        print('Integrated probability, All: %.5f, 2+: %.5f'%(np.sum(fields[:,1]),np.sum(fields[idx,1])))

        slew_readout_time = computeSlewReadoutTime(config_struct, coverage_struct)
        print('Expected time spent on slewing and readout ' + str(slew_readout_time) + ' s.')

        coveragefile = os.path.join(params["outputDir"],'coverage_%s.dat'%telescope)
        fid = open(coveragefile,'w')
        for field in fields:
            fid.write('%d %.10f '%(field[0],field[1]))
            for ii in range(len(filts)):
                 fid.write('%d '%(field[2+ii]))
            fid.write('\n')
        fid.close()

    summaryfile = os.path.join(params["outputDir"],'summary.dat')
    fid = open(summaryfile,'w')

    gpstime = params["gpstime"]
    event_mjd = Time(gpstime, format='gps', scale='utc').mjd

    tts = np.array([1,7,60])
    for tt in tts:

        radecs = np.empty((0,2))
        mjds_floor = []
        mjds = []
        ipixs = np.empty((0,2))
        cum_prob = 0.0
        cum_area = 0.0

        for ii in range(len(coverage_struct["ipix"])):
            data = coverage_struct["data"][ii,:]
            filt = coverage_struct["filters"][ii]
            ipix = coverage_struct["ipix"][ii]
            patch = coverage_struct["patch"][ii]
            FOV = coverage_struct["FOV"][ii]
            area = coverage_struct["area"][ii]

            prob = np.sum(map_struct["prob"][ipix])

            if data[2] > event_mjd+tt:
                continue

            ipixs = np.append(ipixs,ipix)
            ipixs = np.unique(ipixs).astype(int)

            cum_prob = np.sum(map_struct["prob"][ipixs])
            cum_area = len(ipixs) * map_struct["pixarea_deg2"]
            mjds.append(data[2])
            mjds_floor.append(int(np.floor(data[2])))

        if len(mjds_floor) == 0:
            print("No images after %.1f days..."%tt)
            fid.write('%.1f,-1,-1,-1,-1\n'%(tt))
        else:

            mjds = np.unique(mjds)
            mjds_floor = np.unique(mjds_floor)

            print("After %.1f days..."%tt)
            print("Number of hours after first image: %.5f"%(24*(np.min(mjds)-event_mjd)))
            print("MJDs covered: %s"%(" ".join(str(x) for x in mjds_floor)))
            print("Cumultative probability: %.5f"%cum_prob)
            print("Cumultative area: %.5f degrees"%cum_area)

            fid.write('%.1f,%.5f,%.5f,%.5f,%s\n'%(tt,24*(np.min(mjds)-event_mjd),cum_prob,cum_area," ".join(str(x) for x in mjds_floor)))

    fid.close()

def schedule_alternating(params, config_struct, telescope, map_struct, tile_struct, previous_coverage_struct=None):

    if "filt_change_time" in config_struct.keys(): filt_change_time = config_struct["filt_change_time"]
    else: filt_change_time = 0
    if (params["doUpdateScheduler"] or params["doTreasureMap"]) and previous_coverage_struct:
        tile_struct_hold = gwemopt.utils.check_overlapping_tiles(params,tile_struct,previous_coverage_struct) #maps field ids to tile_struct

    filters, exposuretimes = params["filters"], params["exposuretimes"]
    coverage_structs = []
    maxidx = 0

    if params["doBalanceExposure"]: #save tile probabilities
        prob = {}
        for key in tile_struct.keys():
            prob[key] = tile_struct[key]['prob']

    for i in range(len(exposuretimes)):
        params["filters"] = [filters[i]]
        params["exposuretimes"] = [exposuretimes[i]]
        config_struct["exposurelist"] = segments.segmentlist(config_struct["exposurelist"][maxidx:])
        total_nexps  = len(config_struct["exposurelist"])

        # if the duration of a single block is less than 30 min, shift by additional time to add up to 30 min
        if i > 0:
            start = Time(coverage_struct["data"][0][2], format='mjd')
            end =  Time(coverage_struct["data"][-1][2], format='mjd')

            delta = end - start
            delta.format = 'sec'
            duration = delta.value + exposuretimes[i] + filt_change_time
            extra_time = (30 * 60) - duration
            if extra_time > 0: extra_time = extra_time + filt_change_time
            elif extra_time <= 0: extra_time = filt_change_time
            config_struct["exposurelist"] = config_struct["exposurelist"].shift(extra_time / 86400.)
        
        if not params["tilesType"] == "galaxy":
            tile_struct = gwemopt.tiles.powerlaw_tiles_struct(params, config_struct, telescope, map_struct, tile_struct)

        if (params["doUpdateScheduler"] or params["doTreasureMap"]) and previous_coverage_struct: #erases tiles from a previous round
            tile_struct = gwemopt.coverage.update_observed_tiles(params,tile_struct_hold,previous_coverage_struct)

        coverage_struct = gwemopt.scheduler.scheduler(params, config_struct, tile_struct)
        if params["doMaxTiles"]:
            tile_struct,doReschedule = gwemopt.utils.slice_number_tiles(params, telescope, tile_struct, coverage_struct)
            
            # set unbalanced fields to 0
            if params["doBalanceExposure"] and params["unbalanced_tiles"]:
                for key in params["unbalanced_tiles"]:
                    tile_struct[key]['prob'] = 0.0

            if doReschedule:
                coverage_struct = gwemopt.scheduler.scheduler(params, config_struct, tile_struct)

        if len(coverage_struct["exposureused"]) > 0:
            maxidx = int(coverage_struct["exposureused"][-1])
            deltaL = total_nexps - maxidx
        elif len(coverage_struct["exposureused"]) == 0: deltaL = 0

        coverage_structs.append(coverage_struct)

        if params["doBalanceExposure"]: # resets probabilities for each filter block
            for key in tile_struct.keys():
                tile_struct[key]['prob'] = prob[key]

        if deltaL <= 1: break
    params["filters"], params["exposuretimes"] = filters, exposuretimes

    return gwemopt.coverage.combine_coverage_structs(coverage_structs),tile_struct




import os, sys, copy
import numpy as np
import healpy as hp
from astropy.time import Time

from scipy.stats import norm

import matplotlib
#matplotlib.rc('text', usetex=True)
matplotlib.use('Agg')
matplotlib.rcParams.update({'font.size': 16})
matplotlib.rcParams['contour.negative_linestyle'] = 'solid'
import matplotlib.pyplot as plt
from matplotlib.pyplot import cm

from gwemopt.segments import angular_distance
import gwemopt.coverage
import gwemopt.utils

try:
    import ligo.skymap.plot
    cmap = "cylon"
except:
    cmap = 'PuBuGn'

def observability(params,map_struct):

    observability_struct = map_struct["observability"]

    unit='Gravitational-wave probability'
    cbar=False

    for telescope in observability_struct.keys():
        plotName = os.path.join(params["outputDir"],'observability_%s.pdf'%telescope)
        hp.mollview(map_struct["prob"]*observability_struct[telescope]["observability"],title='',unit=unit,cbar=cbar,min=np.min(map_struct["prob"]),max=np.max(map_struct["prob"]),cmap=cmap)
        add_edges()
        plt.show()
        plt.savefig(plotName,dpi=200)
        plt.close('all')

    if params["doMovie"]:
        moviedir = os.path.join(params["outputDir"],'movie')
        if not os.path.isdir(moviedir): os.mkdir(moviedir)
    
        for telescope in observability_struct.keys():
            dts = list(observability_struct[telescope]["dts"].keys())
            dts = np.sort(dts)
            for ii,dt in enumerate(dts):
                plotName = os.path.join(moviedir,'observability-%04d.png'%ii)
                title = "Detectability Map: %.2f Days"%dt
                hp.mollview(map_struct["prob"]*observability_struct[telescope]["dts"][dt],title=title,cbar=cbar,min=np.min(map_struct["prob"]),max=np.max(map_struct["prob"]),cmap=cmap)
                add_edges()
                plt.show()
                plt.savefig(plotName,dpi=200)
                plt.close('all')
    
            moviefiles = os.path.join(moviedir,"observability-%04d.png")
            filename = os.path.join(params["outputDir"],"observability_%s.mpg"%telescope)
            ffmpeg_command = 'ffmpeg -an -y -r 20 -i %s -b:v %s %s'%(moviefiles,'5000k',filename)
            os.system(ffmpeg_command)
            filename = os.path.join(params["outputDir"],"observability_%s.gif"%telescope)
            ffmpeg_command = 'ffmpeg -an -y -r 20 -i %s -b:v %s %s'%(moviefiles,'5000k',filename)
            os.system(ffmpeg_command)
            rm_command = "rm %s/*.png"%(moviedir)
            os.system(rm_command)
    
def tauprob(params,tau,prob):

    plotName = os.path.join(params["outputDir"],'tau_prob.pdf')
    plt.figure()
    plt.plot(tau, prob)
    plt.grid()
    plt.xscale('log')
    plt.yscale('log')
    plt.xlabel('Log of observing time $\tau$',fontsize = 20)
    plt.ylabel('Log of detection prob. given the target is at the observing field',fontsize = 20)
    plt.show()
    plt.savefig(plotName,dpi=200)
    plt.close('all')

def tiles(params,map_struct,tiles_structs):

    unit='Gravitational-wave probability'
    cbar=False

    plotName = os.path.join(params["outputDir"],'tiles.pdf')
    plt.figure()
    hp.mollview(map_struct["prob"],title='',unit=unit,cbar=cbar,cmap=cmap)
    ax = plt.gca()
    for telescope in tiles_structs:
        tiles_struct = tiles_structs[telescope]
        for index in tiles_struct.keys():
            ipix, corners, patch = tiles_struct[index]["ipix"], tiles_struct[index]["corners"], tiles_struct[index]["patch"]
            #hp.visufunc.projplot(corners[:,0], corners[:,1], 'k', lonlat = True)
            if not patch: continue
            hp.projaxes.HpxMollweideAxes.add_patch(ax,patch)
            #tiles.plot()
    add_edges()
    plt.show()
    plt.savefig(plotName,dpi=200)
    plt.close('all')

    if params["doReferences"]:

        rot=(90,0,0)
        plotName = os.path.join(params["outputDir"],'tiles_ref.pdf')
        plt.figure()
        hp.mollview(np.zeros(map_struct["prob"].shape),title='',unit=unit,cbar=cbar,cmap=cmap)
        ax = plt.gca()
        for telescope in tiles_structs:
            config_struct = params["config"][telescope]
            tiles_struct = tiles_structs[telescope]
            if not "reference_images" in config_struct: continue
            for index in tiles_struct.keys():
                if not index in config_struct["reference_images"]: continue
                if len(params["filters"]) == 1:
                    if not params["filters"][0] in config_struct["reference_images"][index]:
                        continue
                ipix, corners, patch = tiles_struct[index]["ipix"], tiles_struct[index]["corners"], tiles_struct[index]["patch"]
                #hp.visufunc.projplot(corners[:,0], corners[:,1], 'k', lonlat = True)
                if not patch: continue
                patch_cpy = copy.copy(patch)
                patch_cpy.axes = None
                patch_cpy.figure = None
                patch_cpy.set_transform(ax.transData)
                hp.projaxes.HpxMollweideAxes.add_patch(ax,patch_cpy)
                #tiles.plot()
        add_edges()
        plt.show()
        plt.savefig(plotName,dpi=200)
        plt.close('all')

def add_edges():

    hp.graticule(verbose=False)
    plt.grid(True)
    lons = np.arange(-150.0,180,30.0)
    lats = np.zeros(lons.shape)
    for lon, lat in zip(lons,lats):
        hp.projtext(lon,lat,"%.0f"%lon,lonlat=True)
    lats = np.arange(-60.0,90,30.0)
    lons = np.zeros(lons.shape)
    for lon, lat in zip(lons,lats):
        hp.projtext(lon,lat,"%.0f"%lat,lonlat=True)

def skymap(params,map_struct):

    unit='Gravitational-wave probability'
    cbar=False

    lons = np.arange(-150.0,180,30.0)
    lats = np.zeros(lons.shape)

    plotName = os.path.join(params["outputDir"],'prob.pdf')
    if np.percentile(map_struct["prob"],99) > 0:
        hp.mollview(map_struct["prob"],title='',unit=unit,cbar=cbar,min=np.percentile(map_struct["prob"],1),max=np.percentile(map_struct["prob"],99),cmap=cmap)
    else:
        hp.mollview(map_struct["prob"],title='',unit=unit,cbar=cbar,min=np.percentile(map_struct["prob"],1),cmap=cmap)
    add_edges()
    plt.show()
    plt.savefig(plotName,dpi=200)
    plt.close('all')

    if "distmu" in map_struct:
        plotName = os.path.join(params["outputDir"],'dist.pdf')
        hp.mollview(map_struct["distmu"],unit='Distance [Mpc]',min=np.percentile(map_struct["distmu"],10),max=np.percentile(map_struct["distmu"],90))
        add_edges()
        plt.show()
        plt.savefig(plotName,dpi=200)
        plt.close('all')

        plotName = os.path.join(params["outputDir"],'dist_median.pdf')
        hp.mollview(map_struct["distmed"],unit='Distance [Mpc]',min=np.percentile(map_struct["distmed"],10),max=np.percentile(map_struct["distmed"],90))
        add_edges()
        plt.show()
        plt.savefig(plotName,dpi=200)
        plt.close('all')

def waw(params, detmaps, t_detmaps, strategy_struct):

    if params["doMovie"]:
        moviedir = os.path.join(params["outputDir"],'movie')
        if not os.path.isdir(moviedir): os.mkdir(moviedir)
        
        for ii in range(len(t_detmaps)):
            t_detmap = t_detmaps[ii]
            detmap = detmaps[ii]

            plotName = os.path.join(moviedir,'detmap-%04d.png'%ii)
            title = "Detectability Map: %.2f Days"%t_detmap
            hp.mollview(detmap,title=title,min=0.0,max=1.0,unit="Probability of Detection")
            add_edges()
            plt.show()
            plt.savefig(plotName,dpi=200)
            plt.close('all')

        moviefiles = os.path.join(moviedir,"detmap-%04d.png")
        filename = os.path.join(params["outputDir"],"detmap.mpg")
        ffmpeg_command = 'ffmpeg -an -y -r 20 -i %s -b:v %s %s'%(moviefiles,'5000k',filename)
        os.system(ffmpeg_command)
        filename = os.path.join(params["outputDir"],"detmap.gif")
        ffmpeg_command = 'ffmpeg -an -y -r 20 -i %s -b:v %s %s'%(moviefiles,'5000k',filename)
        os.system(ffmpeg_command)    
        rm_command = "rm %s/*.png"%(moviedir)
        os.system(rm_command)

def efficiency(params, map_struct, efficiency_structs):

    unit='Gravitational-wave probability'
    cbar=False

    plotName = os.path.join(params["outputDir"],'efficiency.pdf')
    plt.figure()
    for key in efficiency_structs:
        efficiency_struct = efficiency_structs[key]
        plt.plot(efficiency_struct["distances"],efficiency_struct["efficiency"],label=efficiency_struct["legend_label"])
    plt.xlabel('Distance [Mpc]')
    plt.ylabel('Efficiency')
    plt.legend(loc="best")
    plt.ylim([0.01,1])
    plt.show()
    plt.savefig(plotName,dpi=200)
    plt.close('all')

    plotName = os.path.join(params["outputDir"],'injs.pdf')
    plt.figure()
    plt.plot(efficiency_struct["ra"],efficiency_struct["dec"],'kx')
    plt.xlabel('RA [Degrees]')
    plt.ylabel('Declination [Degrees]')
    plt.show()
    plt.savefig(plotName,dpi=200)
    plt.close('all')

    plotName = os.path.join(params["outputDir"],'mollview_injs.pdf')
    plt.figure()
    hp.mollview(map_struct["prob"],unit=unit,cbar=cbar)
    hp.projplot(efficiency_struct["ra"], efficiency_struct["dec"], 'wx', lonlat=True, coord='G')
    add_edges()
    plt.show()
    plt.savefig(plotName,dpi=200)
    plt.close('all')

def coverage(params, map_struct, coverage_struct, catalog_struct=None):

    unit='Gravitational-wave probability'
    cbar=False

    plotName = os.path.join(params["outputDir"],'mollview_coverage.pdf')
    plt.figure()
    hp.mollview(map_struct["prob"],title='',unit=unit,cbar=cbar,cmap=cmap)
    hp.projplot(coverage_struct["data"][:,0], coverage_struct["data"][:,1], 'wx', lonlat=True, coord='G')
    add_edges()
    plt.show()
    plt.savefig(plotName,dpi=200)
    plt.close('all')

    idx = np.isfinite(coverage_struct["data"][:,4])
    if not idx.size: return
    min_time = np.min(coverage_struct["data"][idx,4])
    max_time = np.max(coverage_struct["data"][idx,4])

    plotName = os.path.join(params["outputDir"],'coverage.pdf')
    plt.figure(figsize=(10,8))
    ax = plt.gca()
    for ii in range(len(coverage_struct["ipix"])):
        data = coverage_struct["data"][ii,:]
        filt = coverage_struct["filters"][ii]
        ipix = coverage_struct["ipix"][ii]
        patch = coverage_struct["patch"][ii]
        FOV = coverage_struct["FOV"][ii]

        if filt=="g":
            color = "g"
        elif filt=="r":
            color = "r"
        else:
            color = "k"

        plt.scatter(data[2],data[5],s=20,color=color)

    plt.xlabel("Time [MJD]")
    plt.ylabel("Tile Number")
    plt.show()
    plt.savefig(plotName,dpi=200)
    plt.close('all')

    plotName = os.path.join(params["outputDir"],'tiles_coverage.pdf')
    plt.figure()
    hp.mollview(map_struct["prob"],title='',unit=unit,cbar=cbar, cmap=cmap)
    add_edges()
    ax = plt.gca()
    for ii in range(len(coverage_struct["ipix"])):
        data = coverage_struct["data"][ii,:]
        filt = coverage_struct["filters"][ii]
        ipix = coverage_struct["ipix"][ii]
        patch = coverage_struct["patch"][ii]
        FOV = coverage_struct["FOV"][ii]

        #hp.visufunc.projplot(corners[:,0], corners[:,1], 'k', lonlat = True)
        if patch == []:
            continue

        patch_cpy = copy.copy(patch)
        patch_cpy.axes = None
        patch_cpy.figure = None
        patch_cpy.set_transform(ax.transData)
        hp.projaxes.HpxMollweideAxes.add_patch(ax,patch_cpy)
        #tiles.plot()
    plt.show()
    plt.savefig(plotName,dpi=200)
    plt.close('all')

    diffs = []
    if params["tilesType"] == "galaxy":
        coverage_ras = coverage_struct["data"][:,0]
        coverage_decs = coverage_struct["data"][:,1]
        coverage_mjds = coverage_struct["data"][:,2]

        for ii in range(len(coverage_ras)-1):
            current_ra, current_dec = coverage_ras[ii], coverage_decs[ii]
            current_mjd = coverage_mjds[ii]

            dist = angular_distance(current_ra, current_dec,
                                    coverage_ras[ii+1:],
                                    coverage_decs[ii+1:])
            idx = np.where(dist <= 1/3600.0)[0]
            if len(idx) > 0:
                jj = idx[0]
                diffs.append(np.abs(coverage_struct["data"][ii,2] - coverage_struct["data"][jj,2]))
    else:
        for ii in range(len(coverage_struct["ipix"])):
            ipix = coverage_struct["ipix"][ii]
            for jj in range(len(coverage_struct["ipix"])):
                if ii >= jj: continue
                if coverage_struct["telescope"][ii] == coverage_struct["telescope"][jj]:
                    continue
                ipix2 = coverage_struct["ipix"][jj]
                overlap = np.intersect1d(ipix, ipix2)
                rat = np.array([float(len(overlap)) / float(len(ipix)),
                                float(len(overlap)) / float(len(ipix2))])
                if np.any(rat > 0.5):
                    diffs.append(np.abs(coverage_struct["data"][ii,2] - coverage_struct["data"][jj,2]))

    filename = os.path.join(params["outputDir"],'tiles_coverage_hist.dat')
    fid = open(filename, 'w')
    for ii in range(len(diffs)):
        fid.write('%.10f\n' % diffs[ii])
    fid.close()

    plotName = os.path.join(params["outputDir"],'tiles_coverage_hist.pdf')
    fig = plt.figure(figsize=(12, 8))
    #hist, bin_edges = np.histogram(diffs, bins=20)
    bins = np.linspace(0.0, 24.0, 25)
    plt.hist(24.0*np.array(diffs), bins=bins)
    plt.xlabel('Difference Between Observations [hours]')
    plt.ylabel('Number of Observations')
    plt.show()
    plt.savefig(plotName,dpi=200)
    plt.close('all')

    gpstime = params["gpstime"]
    event_mjd = Time(gpstime, format='gps', scale='utc').mjd

    colors=cm.rainbow(np.linspace(0,1,len(params["telescopes"])))
    plotName = os.path.join(params["outputDir"],'tiles_coverage_int.pdf')

    fig = plt.figure(figsize=(12, 8))

    gs = fig.add_gridspec(4, 1)
    ax1 = fig.add_subplot(gs[0:3, 0], projection='astro hours mollweide')
    ax2 = fig.add_subplot(gs[3, 0])
    ax3 = ax2.twinx()   # mirror them

    plt.axes(ax1)
    hp.mollview(map_struct["prob"],title='',unit=unit,cbar=cbar, cmap=cmap,
                hold=True)
    add_edges()
    ax = plt.gca()
    data = {}

    if params["tilesType"] == "galaxy":
        for telescope, color in zip(params["telescopes"],colors):
            idx = np.where(coverage_struct["telescope"] == telescope)[0]
            hp.projscatter(coverage_struct["data"][idx,0],
                           coverage_struct["data"][idx,1],
                           lonlat=True, 
                           s=10, color=color)
    else:
        for ii in range(len(coverage_struct["ipix"])):
            data = coverage_struct["data"][ii,:]
            filt = coverage_struct["filters"][ii]
            ipix = coverage_struct["ipix"][ii]
            patch = coverage_struct["patch"][ii]
            FOV = coverage_struct["FOV"][ii]

            idx = params["telescopes"].index(coverage_struct["telescope"][ii])
 
            if patch == []:
                continue
            #hp.visufunc.projplot(corners[:,0], corners[:,1], 'k', lonlat = True)
            patch_cpy = copy.copy(patch)
            patch_cpy.axes = None
            patch_cpy.figure = None
            patch_cpy.set_transform(ax.transData)
            patch_cpy.set_facecolor(colors[idx])

            hp.projaxes.HpxMollweideAxes.add_patch(ax,patch_cpy)
            #tiles.plot()

    idxs = np.argsort(coverage_struct["data"][:,2])
    plt.axes(ax2)
    for telescope, color in zip(params["telescopes"],colors):
        ipixs = np.empty((0,2))
        cum_prob = 0.0
        cum_area = 0.0

        tts, cum_probs, cum_areas = [], [], []
        if params["tilesType"] == "galaxy":
            cum_galaxies = []

        for jj, ii in enumerate(idxs):
            if np.mod(jj, 100) == 0:
                print('%s: %d/%d' % (telescope, jj, len(idxs)))

            data = coverage_struct["data"][ii,:]
            filt = coverage_struct["filters"][ii]
            ipix = coverage_struct["ipix"][ii]
            patch = coverage_struct["patch"][ii]
            FOV = coverage_struct["FOV"][ii]
            area = coverage_struct["area"][ii]
            if params["tilesType"] == "galaxy":
                galaxies = coverage_struct["galaxies"][ii] 

            if not telescope == coverage_struct["telescope"][ii]:
                continue

            if params["tilesType"] == "galaxy":
                overlap = np.setdiff1d(galaxies, cum_galaxies)
                if len(overlap) > 0:
                    for galaxy in galaxies:
                        if galaxy in cum_galaxies: continue
                        if catalog_struct is None: continue
                        if params["galaxy_grade"] == "Sloc":
                            cum_prob = cum_prob + catalog_struct["Sloc"][galaxy]
                        elif params["galaxy_grade"] == "S":
                            cum_prob = cum_prob + catalog_struct["S"][galaxy]
                    cum_galaxies = np.append(cum_galaxies,galaxies)
                    cum_galaxies = np.unique(cum_galaxies).astype(int)
                cum_area = len(cum_galaxies)
            else:
                ipixs = np.append(ipixs,ipix)
                ipixs = np.unique(ipixs).astype(int)

                cum_prob = np.sum(map_struct["prob"][ipixs])
                cum_area = len(ipixs) * map_struct["pixarea_deg2"]

            cum_probs.append(cum_prob)
            cum_areas.append(cum_area)
            tts.append(data[2]-event_mjd)

        ax2.plot(tts, cum_probs, color=color, linestyle='-', label=telescope)
        ax3.plot(tts, cum_areas, color=color, linestyle='--') 

    ax2.set_xlabel('Time since event [days]')
    if params["tilesType"] == "galaxy":
        ax2.set_ylabel('Integrated Metric')
    else:
        ax2.set_ylabel('Integrated Probability')

    if params["tilesType"] == "galaxy":
        ax3.set_ylabel('Number of galaxies')
    else:
        ax3.set_ylabel('Sky area [sq. deg.]')

    ipixs = np.empty((0,2))
    cum_prob = 0.0
    cum_area = 0.0

    tts, cum_probs, cum_areas = [], [], []
    if params["tilesType"] == "galaxy":
        cum_galaxies = []

    for jj, ii in enumerate(idxs):
        data = coverage_struct["data"][ii,:]
        filt = coverage_struct["filters"][ii]
        ipix = coverage_struct["ipix"][ii]
        patch = coverage_struct["patch"][ii]
        FOV = coverage_struct["FOV"][ii]
        area = coverage_struct["area"][ii]
        if params["tilesType"] == "galaxy":
            galaxies = coverage_struct["galaxies"][ii]

        if params["tilesType"] == "galaxy":
            overlap = np.setdiff1d(galaxies, cum_galaxies)
            if len(overlap) > 0:
                for galaxy in galaxies:
                    if galaxy in cum_galaxies: continue
                    if catalog_struct is None: continue
                    if params["galaxy_grade"] == "Sloc":
                        cum_prob = cum_prob + catalog_struct["Sloc"][galaxy]
                    elif params["galaxy_grade"] == "S":
                        cum_prob = cum_prob + catalog_struct["S"][galaxy]
                cum_galaxies = np.append(cum_galaxies,galaxies)
                cum_galaxies = np.unique(cum_galaxies).astype(int)
            cum_area = len(cum_galaxies)
        else:
            ipixs = np.append(ipixs,ipix)
            ipixs = np.unique(ipixs).astype(int)

            cum_prob = np.sum(map_struct["prob"][ipixs])
            cum_area = len(ipixs) * map_struct["pixarea_deg2"]

        tts.append(data[2]-event_mjd)
        cum_probs.append(cum_prob)
        cum_areas.append(cum_area)

    ax2.plot(tts, cum_probs, color='k', linestyle='-', label='All')
    ax3.plot(tts, cum_areas, color='k', linestyle='--')

    if len(params["telescopes"]) > 3:
        ax2.legend(loc=1,ncol=3,fontsize=10)
        ax2.set_ylim([0,1])
        ax3.set_ylim([0,2000])
    elif "IRIS" in params["telescopes"]:
        ax2.set_ylim([0,0.3])
        ax3.set_ylim([0,1200])
        ax2.legend(loc=1)     
    elif "ZTF" in params["telescopes"]:
        ax2.set_ylim([0,0.6])
        ax3.set_ylim([0,6000])
        ax2.legend(loc=1)
    elif "PS1" in params["telescopes"]:
        ax2.set_ylim([0,0.6])
        ax3.set_ylim([0,6000])
        ax2.legend(loc=1)
    else:
        ax2.legend(loc=1)
    plt.show()
    plt.savefig(plotName,dpi=200)
    plt.close('all')

    filename = os.path.join(params["outputDir"],'tiles_coverage_int.dat')
    fid = open(filename, 'w')
    for ii in range(len(tts)):
        fid.write('%.10f %.10e %.10f\n' % (tts[ii], cum_probs[ii], cum_areas[ii]))
    fid.close()

    print('Total Cumulative Probability, Area: %.5f, %.5f' % (cum_probs[-1],
                                                              cum_areas[-1]))

    plotName = os.path.join(params["outputDir"],'tiles_coverage_scaled.pdf')
    plt.figure()
    hp.mollview(map_struct["prob"],title='',unit=unit,cbar=cbar,cmap=cmap)
    add_edges()
    ax = plt.gca()
    for ii in range(len(coverage_struct["ipix"])):
        data = coverage_struct["data"][ii,:]
        filt = coverage_struct["filters"][ii]
        ipix = coverage_struct["ipix"][ii]
        patch = coverage_struct["patch"][ii]
        FOV = coverage_struct["FOV"][ii]

        if patch == []:
            continue

        #hp.visufunc.projplot(corners[:,0], corners[:,1], 'k', lonlat = True)
        patch_cpy = copy.copy(patch)
        patch_cpy.axes = None
        patch_cpy.figure = None
        patch_cpy.set_transform(ax.transData)
        current_alpha = patch_cpy.get_alpha()

        if current_alpha > 0.0:
            alpha = data[4]/max_time
            if alpha > 1:
                alpha = 1.0
            patch_cpy.set_alpha(alpha)
        hp.projaxes.HpxMollweideAxes.add_patch(ax,patch_cpy)
        #tiles.plot()
    plt.show()
    plt.savefig(plotName,dpi=200)
    plt.close('all')

    if params["doMovie"]:
        idx = np.isfinite(coverage_struct["data"][:,2])
        mjd_min = np.min(coverage_struct["data"][idx,2])
        mjd_max = np.max(coverage_struct["data"][idx,2])
        mjd_N = 100
    
        mjds = np.linspace(mjd_min,mjd_max,num=mjd_N)
        moviedir = os.path.join(params["outputDir"],'movie')
        if not os.path.isdir(moviedir): os.mkdir(moviedir)
    
        #for jj in range(len(coverage_struct["ipix"])):
        #    mjd = coverage_struct["data"][jj,3]
        for jj in range(len(mjds)):
            mjd = mjds[jj]
            plotName = os.path.join(moviedir,'coverage-%04d.png'%jj)
            title = "Coverage Map: %.2f"%mjd
    
            plt.figure()
            hp.mollview(map_struct["prob"],title=title,unit=unit,cbar=cbar,
                        cmap=cmap)
            add_edges()
            ax = plt.gca()
    
            idx = np.where(coverage_struct["data"][:,2]<=mjd)[0]
            #for ii in range(jj):
            for ii in idx: 
                data = coverage_struct["data"][ii,:]
                filt = coverage_struct["filters"][ii]
                ipix = coverage_struct["ipix"][ii]
                patch = coverage_struct["patch"][ii]
                FOV = coverage_struct["FOV"][ii]
    
                if patch == []:
                    continue

                #hp.visufunc.projplot(corners[:,0], corners[:,1], 'k', lonlat = True)
                patch_cpy = copy.copy(patch)
                patch_cpy.axes = None
                patch_cpy.figure = None
                patch_cpy.set_transform(ax.transData)
                #alpha = data[4]/max_time
                #if alpha > 1:
                #    alpha = 1.0
                #patch_cpy.set_alpha(alpha)
                hp.projaxes.HpxMollweideAxes.add_patch(ax,patch_cpy)
                #tiles.plot()
            plt.show()
            plt.savefig(plotName,dpi=200)
            plt.close('all')
            
        moviefiles = os.path.join(moviedir,"coverage-%04d.png")
        filename = os.path.join(params["outputDir"],"coverage.mpg")
        ffmpeg_command = 'ffmpeg -an -y -r 20 -i %s -b:v %s %s'%(moviefiles,'5000k',filename)
        os.system(ffmpeg_command)
        filename = os.path.join(params["outputDir"],"coverage.gif")
        ffmpeg_command = 'ffmpeg -an -y -r 20 -i %s -b:v %s %s'%(moviefiles,'5000k',filename)
        os.system(ffmpeg_command)
        rm_command = "rm %s/*.png"%(moviedir)
        os.system(rm_command)

def scheduler(params,exposurelist,keys): 
    plotName = os.path.join(params["outputDir"],'scheduler.pdf')
    if params['scheduleType'].endswith('_slew'):
        e = []
        k = []
        start_time = exposurelist[0][0]
        for exposure, key in zip(exposurelist, keys):
            e.append((exposure[0] - start_time) * 24)
            k.append(key)
            e.append((exposure[1] - start_time) * 24)
            k.append(key)
        plt.figure()
        plt.grid()
        plt.xlabel("Time (h)")
        plt.ylabel("Tile Number")
        plt.plot(e, k, 'b-')
    else:
        xs = []
        ys = []
        for ii,key in zip(np.arange(len(exposurelist)),keys):
            xs.append(ii)
            ys.append(key)    
        plt.figure()
        plt.xlabel("Exposure Number")
        plt.ylabel("Tile Number")
        plt.plot(xs,ys,'kx')
    plt.show()
    plt.savefig(plotName,dpi=200)
    plt.close('all')

def transients(params, map_struct, transients_struct):

    unit='Gravitational-wave probability'
    cbar=False

    ra = transients_struct["data"][:,0]
    dec = transients_struct["data"][:,1]

    plotName = os.path.join(params["outputDir"],'transients.pdf')
    plt.figure()
    hp.mollview(map_struct["prob"],unit=unit,cbar=cbar)
    hp.projplot(ra, dec, 'wx', lonlat=True, coord='G')
    add_edges()
    plt.show()
    plt.savefig(plotName,dpi=200)
    plt.close('all')

def doMovie_supersched(params,coverage_structs,tile_structs,map_struct):
    
    unit='Gravitational-wave probability'
    cbar=False
    
    idx = np.isfinite(coverage_structs["data"][:,2])
    try:
        mjd_min = np.min(coverage_structs["data"][idx,2])
    except:
        return
    mjd_max = np.max(coverage_structs["data"][idx,2])
    mjd_N = 100
    Tobs = list(params["Tobs_all"])
    Tobs = np.linspace(Tobs[0],Tobs[1],params["Tobs_split"]+1)
    mjds = np.linspace(mjd_min,mjd_max,num=mjd_N)
    
    parentdir = os.path.abspath(os.path.join(params["outputDir"], os.pardir))
    moviedir = os.path.join(parentdir, 'movie') #saves movie file in parent directory
    
    if not os.path.isdir(moviedir): os.mkdir(moviedir)
    
    #for jj in range(len(coverage_struct["ipix"])):
    #    mjd = coverage_struct["data"][jj,3]
    
    for jj in range(len(mjds)):
        mjd = mjds[jj]
        for i in range(len(Tobs)): #can change this to enumerate(Tobs) if Tobs is variable
            ii = jj+(100*i)
            if not os.path.exists(os.path.join(moviedir,f'coverage-{ii:04d}.png')): #adds multiples of 100 for each round of Tobs
                plotName = os.path.join(moviedir,f'coverage-{ii:04d}.png')
                break
    
        title = "Coverage Map: %.2f"%mjd
        plt.figure()
        hp.mollview(map_struct["prob"],title=title,unit=unit,cbar=cbar,cmap=cmap)
        add_edges()
        ax = plt.gca()
        
        idx = np.where(coverage_structs["data"][:,2]<=mjd)[0]
        
        #for ii in range(jj):
        for ii in idx:
            data = coverage_structs["data"][ii,:]
            filt = coverage_structs["filters"][ii]
            ipix = coverage_structs["ipix"][ii]
            patch = coverage_structs["patch"][ii]
            FOV = coverage_structs["FOV"][ii]
            
            if patch == []:
                continue
        
            #hp.visufunc.projplot(corners[:,0], corners[:,1], 'k', lonlat = True)
            patch_cpy = copy.copy(patch)
            patch_cpy.axes = None
            patch_cpy.figure = None
            patch_cpy.set_transform(ax.transData)
            
            #alpha = data[4]/max_time
            #if alpha > 1:
            #    alpha = 1.0
            #patch_cpy.set_alpha(alpha)
            hp.projaxes.HpxMollweideAxes.add_patch(ax,patch_cpy)
        #tiles.plot()

        if len(params["coverage_structs"]) == 1:
            continue
        else:
            i = len(params["coverage_structs"])-1 #finds out which round we are in

        while i>0: #goes through all previous rounds
            readfile = os.path.join(parentdir,f'{Tobs[i-1]:.2}_to_{Tobs[i]:.2}_Tobs')
            prevtelescopes = params["alltelescopes"][i-1].split(",")
            prev_tile_structs = params["tile_structs"][f'tile_structs_{i-1}']
            i-=1
            for prevtelescope in prevtelescopes:
                if prevtelescope not in prev_tile_structs: continue
                prev_tile_struct = prev_tile_structs[prevtelescope]
                schedfile = f'schedule_{prevtelescope}.dat'
                data_file = os.path.join(readfile,schedfile)
            
                with open(data_file, "r") as f:
                    for line in f:
                        data = list(line.split(' '))
                        field_id = int(data[0])
                        if int(data[8]) == 1:
                            patch = prev_tile_struct[field_id]["patch"]
                            if patch == []:
                                continue
                            patch_cpy = copy.copy(patch)
                            patch_cpy.axes = None
                            patch_cpy.figure = None
                            patch_cpy.set_transform(ax.transData)
                            patch_cpy.set_facecolor('white')
                            hp.projaxes.HpxMollweideAxes.add_patch(ax,patch_cpy)
    
        plt.show()
        plt.savefig(plotName,dpi=200)
        plt.close('all')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import copy
import h5py
import numpy as np
import healpy as hp

from scipy.stats import norm
from scipy.special import gammaincinv

from astroquery.vizier import Vizier
from ligo.skymap import distance
from astropy.table import Table
from astropy.io import ascii
from astropy.cosmology import WMAP9 as cosmo
from astropy.table import Column
import astropy.units as u
import scipy.stats
# Unset row limits when querying Vizier
Vizier.ROW_LIMIT = -1


def get_catalog(params, map_struct):
    if not os.path.isdir(params["catalogDir"]):
        os.makedirs(params["catalogDir"])

    catalogFile = os.path.join(params["catalogDir"],
                               "%s.hdf5" % params["galaxy_catalog"])

    """AB Magnitude zero point."""
    MAB0 = -2.5 * np.log10(3631.e-23)
    pc_cm = 3.08568025e18
    const = 4. * np.pi * (10. * pc_cm)**2.

    if params["galaxy_catalog"] == "2MRS":
        if not os.path.isfile(catalogFile):
            import astropy.constants as c

            cat, = Vizier.get_catalogs('J/ApJS/199/26/table3')

            ra, dec = cat["RAJ2000"], cat["DEJ2000"]
            cz = cat["cz"]
            magk = cat["Ktmag"]

            z = (u.Quantity(cat['cz']) / c.c).to(u.dimensionless_unscaled)

            completeness = 0.5
            alpha = -1.0
            MK_star = -23.55
            MK_max = MK_star + 2.5 * np.log10(gammaincinv(alpha + 2,
                                                          completeness))
            MK = magk - cosmo.distmod(z)
            idx = (z > 0) & (MK < MK_max)

            ra, dec = ra[idx], dec[idx]
            z = z[idx]
            magk = magk[idx]

            distmpc = cosmo.luminosity_distance(z).to('Mpc').value

            with h5py.File(catalogFile, 'w') as f:
                f.create_dataset('ra', data=ra)
                f.create_dataset('dec', data=dec)
                f.create_dataset('z', data=z)
                f.create_dataset('magk', data=magk)
                f.create_dataset('distmpc', data=distmpc)

        else:
            with h5py.File(catalogFile, 'r') as f:
                ra, dec = f['ra'][:], f['dec'][:]
                z = f['z'][:]
                magk = f['magk'][:]
                distmpc = f['distmpc'][:]
        r = distmpc * 1.0
        mag = magk * 1.0

    elif params["galaxy_catalog"] == "GLADE":
        if not os.path.isfile(catalogFile):
            cat, = Vizier.get_catalogs('VII/281/glade2')

            ra, dec = cat["RAJ2000"], cat["DEJ2000"]
            distmpc, z = cat["Dist"], cat["z"]
            magb, magk = cat["Bmag"], cat["Kmag"]
            # Keep track of galaxy identifier
            GWGC, PGC, HyperLEDA = cat["GWGC"], cat["PGC"], cat["HyperLEDA"]
            _2MASS, SDSS = cat["_2MASS"], cat["SDSS-DR12"]

            idx = np.where(distmpc >= 0)[0]
            ra, dec = ra[idx], dec[idx]
            distmpc, z = distmpc[idx], z[idx]
            magb, magk = magb[idx], magk[idx]
            GWGC, PGC, HyperLEDA = GWGC[idx], PGC[idx], HyperLEDA[idx]
            _2MASS, SDSS = _2MASS[idx], SDSS[idx]

            
            with h5py.File(catalogFile, 'w') as f:
                f.create_dataset('ra', data=ra)
                f.create_dataset('dec', data=dec)
                f.create_dataset('distmpc', data=distmpc)
                f.create_dataset('magb', data=magb)
                f.create_dataset('magk', data=magk)
                f.create_dataset('z', data=z)
                # Add galaxy identifier
                f.create_dataset('GWGC', data=GWGC)
                f.create_dataset('PGC', data=PGC)
                f.create_dataset('HyperLEDA', data=HyperLEDA)
                f.create_dataset('2MASS', data=_2MASS)
                f.create_dataset('SDSS', data=SDSS)

        else:
            with h5py.File(catalogFile, 'r') as f:
                ra, dec = f['ra'][:], f['dec'][:]
                distmpc, z = f['distmpc'][:], f['z'][:]
                magb, magk = f['magb'][:], f['magk'][:]
                GWGC, PGC, _2MASS = f['GWGC'][:], f['PGC'][:], f['2MASS'][:]
                HyperLEDA, SDSS = f['HyperLEDA'][:], f['SDSS'][:]
                # Convert bytestring to unicode
                GWGC = GWGC.astype('U')
                PGC = PGC.astype('U')
                HyperLEDA = HyperLEDA.astype('U')
                _2MASS = _2MASS.astype('U')
                SDSS = SDSS.astype('U')

        # Keep only galaxies with finite B mag when using it in the grade
        if params["galaxy_grade"] == "S":
            idx = np.where(~np.isnan(magb))[0]
            ra, dec, distmpc = ra[idx], dec[idx], distmpc[idx]
            magb, magk = magb[idx], magk[idx]
            GWGC, PGC, HyperLEDA = GWGC[idx], PGC[idx], HyperLEDA[idx]
            _2MASS, SDSS = _2MASS[idx], SDSS[idx]

        r = distmpc * 1.0
        mag = magb * 1.0

    elif params["galaxy_catalog"] == "CLU":
        if not os.path.isfile(catalogFile):
            raise ValueError("Please add %s." % catalogFile)

        cat = Table.read(catalogFile)
        name = cat['name']
        ra, dec = cat['ra'], cat['dec']
        sfr_fuv, mstar = cat['sfr_fuv'], cat['mstar']
        distmpc, magb = cat['distmpc'], cat['magb']
        a, b2a, pa = cat['a'], cat['b2a'], cat['pa']
        btc = cat['btc']

        idx = np.where(distmpc >= 0)[0]
        ra, dec = ra[idx], dec[idx]
        sfr_fuv, mstar = sfr_fuv[idx], mstar[idx]
        distmpc, magb = distmpc[idx], magb[idx]
        a, b2a, pa = a[idx], b2a[idx], pa[idx]
        btc = btc[idx]

        idx = np.where(~np.isnan(magb))[0]
        ra, dec = ra[idx], dec[idx]
        sfr_fuv, mstar = sfr_fuv[idx], mstar[idx]
        distmpc, magb = distmpc[idx], magb[idx]
        a, b2a, pa = a[idx], b2a[idx], pa[idx]
        btc = btc[idx]

        z = -1*np.ones(distmpc.shape)
        r = distmpc * 1.0
        mag = magb * 1.0


    elif params["galaxy_catalog"] == "mangrove":
        catalogFile = os.path.join(params["catalogDir"],
                               "%s.hdf5" % params["galaxy_catalog"])
        
        if not os.path.isfile(catalogFile):
            print("mangrove catalog not found localy, start the automatic download")
            url = 'https://mangrove.lal.in2p3.fr/data/mangrove.hdf5'
            os.system("wget -O {}/mangrove.hdf5 {}".format(params["catalogDir"], url))
            
        cat = Table.read(catalogFile)
        
        ra, dec = cat["RA"], cat["dec"]
        distmpc, z = cat["dist"], cat["z"]
        magb, magk = cat["B_mag"], cat["K_mag"]
        magW1, magW2, magW3, magW4 = cat["w1mpro"], cat["w2mpro"], cat["w3mpro"], cat["w4mpro"]
        stellarmass = cat['stellarmass']
        # Keep track of galaxy identifier
        GWGC, PGC, HyperLEDA = cat["GWGC_name"], cat["PGC"], cat["HyperLEDA_name"]
        _2MASS, SDSS = cat["2MASS_name"], cat["SDSS-DR12_name"]

        idx = np.where(distmpc >= 0)[0]
        ra, dec = ra[idx], dec[idx]
        distmpc, z = distmpc[idx], z[idx]
        magb, magk = magb[idx], magk[idx]
        magW1, magW2, magW3, magW4 = magW1[idx], magW2[idx], magW3[idx], magW4[idx]
        stellarmass = stellarmass[idx]
        GWGC, PGC, HyperLEDA = GWGC[idx], PGC[idx], HyperLEDA[idx]
        _2MASS, SDSS = _2MASS[idx], SDSS[idx]
        

        # Convert bytestring to unicode
        GWGC = GWGC.astype('U')
        PGC = PGC.astype('U')
        HyperLEDA = HyperLEDA.astype('U')
        _2MASS = _2MASS.astype('U')
        SDSS = SDSS.astype('U')

        r = distmpc * 1.0
        mag = magb * 1.0

    n, cl = params["powerlaw_n"], params["powerlaw_cl"]
    dist_exp = params["powerlaw_dist_exp"]

    prob_scaled = copy.deepcopy(map_struct["prob"])
    prob_sorted = np.sort(prob_scaled)[::-1]
    prob_indexes = np.argsort(prob_scaled)[::-1]
    prob_cumsum = np.cumsum(prob_sorted)
    index = np.argmin(np.abs(prob_cumsum - cl)) + 1
    prob_scaled[prob_indexes[index:]] = 0.0
    prob_scaled = prob_scaled**n

    theta = 0.5 * np.pi - dec * 2 * np.pi / 360.0
    phi = ra * 2 * np.pi / 360.0
    ipix = hp.ang2pix(map_struct["nside"], ra, dec, lonlat=True)

    if "distnorm" in map_struct:
        if map_struct["distnorm"] is not None:
            
            #creat an mask to cut at 3 sigma in distance
            mask = np.zeros(len(r))
            
            #calculate the moments from distmu, distsigma and distnorm
            mom_mean, mom_std, mom_norm = distance.parameters_to_moments(map_struct["distmu"],map_struct["distsigma"])

            condition_indexer = np.where( (r < (mom_mean[ipix] + (3*mom_std[ipix]))) & (r > (mom_mean[ipix] - (3*mom_std[ipix])) )) 
            mask[condition_indexer] = 1

            Sloc = prob_scaled[ipix] * (map_struct["distnorm"][ipix] *
                                        norm(map_struct["distmu"][ipix],
                                        map_struct["distsigma"][ipix]).pdf(r))**params["powerlaw_dist_exp"] / map_struct["pixarea"]
    
            #multiplie the Sloc by 1 or 0 according to the 3 sigma condistion
            Sloc = Sloc*mask
            idx = np.where(condition_indexer)[0]
        else:
            Sloc = copy.copy(prob_scaled[ipix])
            idx = np.arange(len(r)).astype(int)
    else:
        Sloc = copy.copy(prob_scaled[ipix])
        idx = np.arange(len(r)).astype(int)

    # this happens when we are using a tiny catalog...
    if np.all(Sloc == 0.0):
        Sloc[:] = 1.0


    #new version of the Slum calcul (from HOGWARTs)
    Lsun = 3.828e26
    Msun = 4.83
    Mknmin = -19
    Mknmax = -12
    Lblist = []
    
    for i in range(0, len(r)):
        Mb = mag[i] - 5 * np.log10((r[i] * 10 ** 6)) + 5
        #L = Lsun * 2.512 ** (Msun - Mb)
        Lb = Lsun * 2.512 ** (Msun - Mb)
        Lblist.append(Lb)
        
                           
    #set 0 when Sloc is 0 (keep compatible galaxies for normalization)
    Lblist = np.array(Lblist)
    
    Lblist[Sloc == 0] = 0

    Slum = Lblist / np.nansum(np.array(Lblist))
   
  
    """
    L_nu = const * 10.**((mag + MAB0)/(-2.5))
    L_nu = L_nu / np.nanmax(L_nu[idx])
    L_nu = L_nu**params["catalog_n"]
    L_nu[L_nu < 0.001] = 0.001
    L_nu[L_nu > 1.0] = 1.0
    Slum = L_nu / np.sum(L_nu)
    """


    mlim, M_KNmin, M_KNmax = 22, -17, -12
    L_KNmin = const * 10.**((M_KNmin + MAB0)/(-2.5))
    L_KNmax = const * 10.**((M_KNmax + MAB0)/(-2.5))

    Llim = 4. * np.pi * (r * 1e6 * pc_cm)**2. * 10.**((mlim + MAB0)/(-2.5))
    Sdet = (L_KNmax-Llim)/(L_KNmax-L_KNmin)
    Sdet[Sdet < 0.01] = 0.01
    Sdet[Sdet > 1.0] = 1.0

    # Set nan values to zero
    Sloc[np.isnan(Sloc)] = 0
    Slum[np.isnan(Slum)] = 0

    if params["galaxy_grade"] == "Smass":
        
        if params["galaxy_catalog"] != "mangrove":
            raise ValueError("You are trying to use the stellar mass information (Smass), please select the mangrove catalog for such use.")

        #set Smass
        Smass = np.array(stellarmass)
        
        #put null values to nan
        Smass[np.where(Smass == 0)] = np.nan

        #go back to linear scaling and not log
        Smass = 10**Smass

        # Keep only galaxies with finite stellarmass when using it in the grade
        #set nan values to 0
        Smass[~np.isfinite(Smass)] = 0  
 
        #set Smass
        Smass = Smass / np.sum(Smass)
        
	#alpha is defined only with non null mass galaxies, we set a mask for that
        ind_without_mass = np.where(Smass == 0)
        Sloc_temp = copy.deepcopy(Sloc)
        Sloc_temp[ind_without_mass] = 0

        #alpha_mass parameter is defined in such way that in mean Sloc count in as much as Sloc*alpha*Smass
        alpha_mass = (np.sum(Sloc_temp) / np.sum( Sloc_temp*Smass ) )
        print("You chose to use the grade using stellar mass, the parameters values are:")
        print("alpha_mass =", alpha_mass)

	#beta_mass is a parameter allowing to change the importance of Sloc according to Sloc*alpha*Smass
        #beta_mass should be fitted in the futur on real GW event for which we have the host galaxy
        #fixed to one at the moment
        beta_mass = 1
        print("beta_mass =", beta_mass)

        Smass = Sloc*(1+ (alpha_mass*beta_mass*Smass) )
        #Smass = Sloc*Smass
        
    S = Sloc*Slum*Sdet

    prob = np.zeros(map_struct["prob"].shape)
    if params["galaxy_grade"] == "Sloc":     
        for j in range(len(ipix)):
            prob[ipix[j]] += Sloc[j]
        grade = Sloc
    elif params["galaxy_grade"] == "S":
        for j in range(len(ipix)):
            prob[ipix[j]] += S[j]
        grade = S
    elif params["galaxy_grade"] == "Smass":
        for j in range(len(ipix)):
            prob[ipix[j]] += Smass[j]
        grade = Smass
 
    prob = prob / np.sum(prob)

    map_struct['prob_catalog'] = prob
    if params["doUseCatalog"]:
        map_struct['prob'] = prob

    Lblist = np.array(Lblist)

    idx = np.where(~np.isnan(grade))[0]
    grade = grade[idx]
    ra, dec, Sloc, S = ra[idx], dec[idx], Sloc[idx], S[idx]
    distmpc, z = distmpc[idx], z[idx]

    if params["galaxy_catalog"] == "GLADE":
        GWGC, PGC, HyperLEDA = GWGC[idx], PGC[idx], HyperLEDA[idx]
        _2MASS, SDSS = _2MASS[idx], SDSS[idx]
        Lblist = Lblist[idx]
        magb = magb[idx]
 
    if params["galaxy_catalog"] == "mangrove":
        GWGC, PGC, HyperLEDA = GWGC[idx], PGC[idx], HyperLEDA[idx]
        _2MASS, SDSS = _2MASS[idx], SDSS[idx]   
        stellarmass, magb = stellarmass[idx], magb[idx]
        if params["galaxy_grade"] == "Smass":
            Smass = Smass[idx]

    """
    Sthresh = np.max(grade)*0.01
    idx = np.where(grade >= Sthresh)[0]
    grade = grade[idx]
    ra, dec, Sloc, S = ra[idx], dec[idx], Sloc[idx], S[idx]
    distmpc, z = distmpc[idx], z[idx]
    if params["galaxy_catalog"] == "GLADE":
        GWGC, PGC, HyperLEDA = GWGC[idx], PGC[idx], HyperLEDA[idx]
        _2MASS, SDSS = _2MASS[idx], SDSS[idx]
    """
    
    idx = np.argsort(grade)[::-1]
    grade = grade[idx]

    ra, dec, Sloc, S = ra[idx], dec[idx], Sloc[idx], S[idx]
    distmpc, z = distmpc[idx], z[idx]
    if params["galaxy_catalog"] == "GLADE":
        GWGC, PGC, HyperLEDA = GWGC[idx], PGC[idx], HyperLEDA[idx]
        _2MASS, SDSS = _2MASS[idx], SDSS[idx]
        Lblist = Lblist[idx]
        magb = magb[idx]

    if params["galaxy_catalog"] == "mangrove":
        GWGC, PGC, HyperLEDA = GWGC[idx], PGC[idx], HyperLEDA[idx]
        _2MASS, SDSS = _2MASS[idx], SDSS[idx]
        stellarmass, magb = stellarmass[idx], magb[idx]
        if params["galaxy_grade"] == "Smass":
            Smass = Smass[idx]
       
    # Keep only galaxies within 3sigma in distance
    if params["galaxy_catalog"] != "mangrove":
        mask = Sloc > 0
        ra, dec, Sloc, S = ra[mask], dec[mask], Sloc[mask], S[mask],
        distmpc, z = distmpc[mask], z[mask]
        if params["galaxy_catalog"] == "GLADE":
            GWGC, PGC, HyperLEDA = GWGC[mask], PGC[mask], HyperLEDA[mask]
            _2MASS, SDSS = _2MASS[mask], SDSS[mask]
            Lblist = Lblist[mask]

    if params["galaxy_catalog"] == "mangrove":
        # Keep only galaxies within 3sigma in distance
        mask = Sloc > 0
        ra, dec, Sloc, S = ra[mask], dec[mask], Sloc[mask], S[mask]
        distmpc, z = distmpc[mask], z[mask]
        GWGC, PGC, HyperLEDA = GWGC[mask], PGC[mask], HyperLEDA[mask]
        _2MASS, SDSS = _2MASS[mask], SDSS[mask]
        stellarmass, magb = stellarmass[mask], magb[mask]

        if params["galaxy_grade"] == "Smass":
            Smass = Smass[mask]

        
    if len(ra) > 2000:
        print('Cutting catalog to top 2000 galaxies...')
        idx = np.arange(2000).astype(int)
        ra, dec, Sloc, S = ra[idx], dec[idx], Sloc[idx], S[idx]
        distmpc, z = distmpc[idx], z[idx]
        if params["galaxy_catalog"] == "GLADE":
            GWGC, PGC, HyperLEDA = GWGC[idx], PGC[idx], HyperLEDA[idx]
            _2MASS, SDSS = _2MASS[idx], SDSS[idx]
            Lblist = Lblist[idx]
            magb = magb[idx]

        elif params["galaxy_catalog"] == "mangrove":
            GWGC, PGC, HyperLEDA = GWGC[idx], PGC[idx], HyperLEDA[idx]
            _2MASS, SDSS = _2MASS[idx], SDSS[idx]
            stellarmass, magb = stellarmass[idx], magb[idx]

            if params["galaxy_grade"] == "Smass":
                Smass = Smass[idx]

    # now normalize the distributions
    S = S / np.sum(S)
    Sloc = Sloc / np.sum(Sloc)

    if params["galaxy_grade"] == "Smass":
        Smass = Smass/np.sum(Smass)
    else:
        Smass = np.ones(Sloc.shape)
        Smass = Smass/np.sum(Smass)

    catalog_struct = {}
    catalog_struct["ra"] = ra
    catalog_struct["dec"] = dec
    catalog_struct["Sloc"] = Sloc
    catalog_struct["S"] = S
    catalog_struct["Smass"] = Smass

    if params["writeCatalog"]:
        catalogfile = os.path.join(params["outputDir"], 'catalog.csv')
        fid = open(catalogfile, 'w')
        cnt = 1
        if params["galaxy_catalog"] == "GLADE":

            if params["galaxy_grade"] == "S":
 
                fid.write("id, RAJ2000, DEJ2000, Sloc, S, Dist, z,GWGC, PGC, HyperLEDA, 2MASS, SDSS, BLum\n")
                for a, b, c, d, e, f, g, h, i, j, k, l in zip(ra, dec, Sloc, S, distmpc, z, GWGC, PGC, HyperLEDA, _2MASS, SDSS, Lblist):
                    fid.write("%d, %.5f, %.5f, %.5e, %.5e, %.4f, %.4f, %s, %s, %s, %s, %s, %.2E\n" % (cnt, a, b, c, d, e, f, g, h, i, j, k, l))
                    cnt = cnt + 1
                
            else:

                fid.write("id, RAJ2000, DEJ2000, Sloc, S, Dist, z, GWGC, PGC, HyperLEDA, 2MASS, SDSS\n")
                for a, b, c, d, e, f, g, h, i, j, k in zip(ra, dec, Sloc, S, distmpc, z, GWGC, PGC, HyperLEDA, _2MASS, SDSS):
                    fid.write("%d, %.5f, %.5f, %.5e, %.5e, %.4f, %.4f, %s, %s, %s, %s, %s\n" % (cnt, a, b, c, d, e, f, g, h, i, j, k))
                    cnt = cnt + 1

        elif params["galaxy_catalog"] == "mangrove":

            if params["galaxy_grade"] == "Smass":
                fid.write("id, RAJ2000, DEJ2000, Smass, S, Sloc, Dist, z, GWGC, PGC, HyperLEDA, 2MASS, SDSS, B_mag, stellarmass\n")
                for a, b, c, d, e, f, g, h, i, j, k, l, m, n in zip(ra, dec, Smass, S, Sloc, distmpc, z, GWGC, PGC, HyperLEDA, _2MASS, SDSS, magb, stellarmass):
                    fid.write("%d, %.5f, %.5f, %.5e, %.5e, %.5e, %.4f, %.4f, %s, %s, %s, %s, %s, %s, %s\n" % (cnt, a, b, c, d, e, f, g, h, i, j, k, l, m, n))
                    cnt = cnt + 1
            else:
                fid.write("id, RAJ2000, DEJ2000, Sloc, S, Dist, z, GWGC, PGC, HyperLEDA, 2MASS, SDSS\n")
                for a, b, c, d, e, f, g, h, i, j, k in zip(ra, dec, Sloc, S, distmpc, z, GWGC, PGC, HyperLEDA, _2MASS, SDSS):
                    fid.write("%d, %.5f, %.5f, %.5e, %.5e, %.4f, %.4f, %s, %s, %s, %s, %s\n" % (cnt, a, b, c, d, e, f, g, h, i, j, k))
                    cnt = cnt + 1
                
        else:
            fid.write("id, RAJ2000, DEJ2000, Sloc, S, Dist, z\n")
            for a, b, c, d in zip(ra, dec, Sloc, S, distmpc, z):
                fid.write("%d, %.5f, %.5f, %.5e, %.5e, %.4f, %.4f\n" % (cnt, a, b, c, d, e, f))
                cnt = cnt + 1

        fid.close()

    return map_struct, catalog_struct





import os, sys
import copy
import numpy as np
import healpy as hp
import gwemopt.plotting
from astropy.time import Time
from astropy.coordinates import get_sun, SkyCoord
from astropy import units as u

import gwemopt.utils, gwemopt.tiles
import gwemopt.rankedTilesGenerator
import gwemopt.scheduler
import ligo.segments as segments

def combine_coverage_structs(coverage_structs):

    coverage_struct_combined = {}
    coverage_struct_combined["data"] = np.empty((0,8))
    coverage_struct_combined["filters"] = np.empty((0,1))
    coverage_struct_combined["ipix"] = []
    coverage_struct_combined["patch"] = []
    coverage_struct_combined["FOV"] = np.empty((0,1))
    coverage_struct_combined["area"] = np.empty((0,1))
    coverage_struct_combined["telescope"] = np.empty((0,1))
    coverage_struct_combined["galaxies"] = []    

    for coverage_struct in coverage_structs:
        coverage_struct_combined["data"] = np.append(coverage_struct_combined["data"],coverage_struct["data"],axis=0)
        coverage_struct_combined["filters"] = np.append(coverage_struct_combined["filters"],coverage_struct["filters"])
        coverage_struct_combined["ipix"] = coverage_struct_combined["ipix"] + coverage_struct["ipix"]
        coverage_struct_combined["patch"] = coverage_struct_combined["patch"] + coverage_struct["patch"]
        coverage_struct_combined["FOV"] = np.append(coverage_struct_combined["FOV"],coverage_struct["FOV"])
        coverage_struct_combined["area"] = np.append(coverage_struct_combined["area"],coverage_struct["area"])
        coverage_struct_combined["telescope"] = np.append(coverage_struct_combined["telescope"],coverage_struct["telescope"])
        if "galaxies" in coverage_struct:
            coverage_struct_combined["galaxies"] = coverage_struct_combined["galaxies"] + coverage_struct["galaxies"]

    return coverage_struct_combined

def read_coverage(params, telescope, filename):

    nside = params["nside"]
    config_struct = params["config"][telescope]

    lines = [line.rstrip('\n') for line in open(filename)]
    lines = lines[1:]
    lines = filter(None,lines)

    coverage_struct = {}
    coverage_struct["data"] = np.empty((0,8))
    coverage_struct["filters"] = []
    coverage_struct["ipix"] = []
    coverage_struct["patch"] = []
    coverage_struct["area"] = []

    for line in lines:
        lineSplit = line.split(",")
        ra = float(lineSplit[2])
        dec = float(lineSplit[3])
        mjd = float(lineSplit[4])
        filt = lineSplit[6]
        mag = float(lineSplit[7])

        coverage_struct["data"] = np.append(coverage_struct["data"],np.array([[ra,dec,mjd,mag,config_struct["exposuretime"],-1,-1,-1]]),axis=0)
        coverage_struct["filters"].append(filt)

        if telescope == "ATLAS":
            alpha=0.2
            color='#6c71c4'
        elif telescope == "PS1":
            alpha=0.1
            color='#859900'
        else:
            alpha=0.2
            color='#6c71c4'

        if config_struct["FOV_coverage_type"] == "square":
            ipix, radecs, patch, area = gwemopt.utils.getSquarePixels(ra, dec, config_struct["FOV_coverage"], nside, alpha=alpha, color=color)
        elif config_struct["FOV_coverage_type"] == "circle":
            ipix, radecs, patch, area = gwemopt.utils.getCirclePixels(ra, dec, config_struct["FOV_coverage"], nside, alpha=alpha, color=color)

        coverage_struct["patch"].append(patch)
        coverage_struct["ipix"].append(ipix)
        coverage_struct["area"].append(area)

    coverage_struct["filters"] = np.array(coverage_struct["filters"])
    coverage_struct["area"] = np.array(coverage_struct["area"])
    coverage_struct["FOV"] = config_struct["FOV_coverage"]*np.ones((len(coverage_struct["filters"]),))

    return coverage_struct

def read_coverage_files(params):

    coverage_structs = []
    for telescope, coverageFile in zip(params["telescopes"],params["coverageFiles"]):
        coverage_struct = read_coverage(params,telescope,coverageFile)
        coverage_structs.append(coverage_struct)

    return combine_coverage_structs(coverage_structs)

def waw(params, map_struct, tile_structs): 

    nside = params["nside"]

    t = np.arange(0,7,1/24.0)
    #t = np.arange(0,7,1.0)
    cr90 = map_struct["cumprob"] < 0.9
    detmaps = gwemopt.waw.detectability_maps(params, t, map_struct, verbose=True, limit_to_region=cr90, nside=nside)

    coverage_structs = []
    for telescope in params["telescopes"]: 
        tile_struct = tile_structs[telescope]
        config_struct = params["config"][telescope]
        T_int = config_struct["exposuretime"]
        ranked_tile_probs = gwemopt.tiles.compute_tiles_map(tile_struct, map_struct["prob"], func='np.sum(x)')
        strategy_struct = gwemopt.waw.construct_followup_strategy_tiles(map_struct["prob"],detmaps,t,tile_struct,T_int,params["Tobs"])
        if strategy_struct is None:
            raise ValueError("Change distance scale...")
        strategy_struct = strategy_struct*86400.0
        keys = tile_struct.keys()
        for key, prob, exposureTime in zip(keys, ranked_tile_probs, strategy_struct):
            tile_struct[key]["prob"] = prob
            tile_struct[key]["exposureTime"] = exposureTime
            tile_struct[key]["nexposures"] = int(np.floor(exposureTime/config_struct["exposuretime"]))
        coverage_struct = gwemopt.scheduler.scheduler(params, config_struct, tile_struct)
        coverage_structs.append(coverage_struct)

    if params["doPlots"]:
        gwemopt.plotting.waw(params,detmaps,t,strategy_struct)

    return combine_coverage_structs(coverage_structs)

def powerlaw(params, map_struct, tile_structs,previous_coverage_struct=None):

    map_struct_hold = copy.deepcopy(map_struct)

    coverage_structs = []
    n_scope = 0
    full_prob_map = map_struct["prob"]

    for jj, telescope in enumerate(params["telescopes"]):

        if params["doSplit"]:
            if "observability" in map_struct:
                map_struct["observability"][telescope]["prob"] = map_struct["groups"][n_scope]
            else:
                map_struct["prob"] = map_struct["groups"][n_scope]
            if n_scope < len(map_struct["groups"]) - 1:
                n_scope += 1
            else:
                n_scope = 0

        config_struct = params["config"][telescope]
        tile_struct = tile_structs[telescope]

        # Try to load the minimum duration of time from telescope config file
        # Otherwise set it to zero
        try:   
            min_obs_duration = config_struct["min_observability_duration"] / 24
        except:
            min_obs_duration = 0.0

        if "filt_change_time" in config_struct.keys(): filt_change_time = config_struct["filt_change_time"]
        else: filt_change_time = 0

        if params["doIterativeTiling"] and (params["tilesType"] == "galaxy"):
            tile_struct = gwemopt.utils.slice_galaxy_tiles(params, tile_struct, combine_coverage_structs(coverage_structs))

        if params["doPerturbativeTiling"] and (jj>0) and (not params["tilesType"] == "galaxy"):
            tile_struct = gwemopt.utils.perturb_tiles(params, config_struct, telescope, map_struct_hold, tile_struct)
 
        if params["doOverlappingScheduling"]:
            tile_struct = gwemopt.utils.check_overlapping_tiles(params, tile_struct, combine_coverage_structs(coverage_structs))

        if params["doAlternatingFilters"]:
            if params["doBlocks"]:
                tile_struct = gwemopt.utils.eject_tiles(params,telescope,tile_struct)
                   
            params_hold = copy.copy(params)
            config_struct_hold = copy.copy(config_struct)
            
            coverage_struct,tile_struct = gwemopt.scheduler.schedule_alternating(params_hold, config_struct_hold, telescope, map_struct_hold, tile_struct,previous_coverage_struct)

            if params["doBalanceExposure"]:
                
                optimized_bool = False
                if not params["doMaxTiles"]: #optimize max tiles (iff max tiles not already specified)
                    optimized_bool = True
                    optimized_max = gwemopt.utils.optimize_max_tiles(params,tile_struct,coverage_struct,config_struct,telescope,map_struct_hold)
                    params["max_nb_tiles"] = np.array([optimized_max],dtype=np.float)

                params_hold = copy.copy(params)
                config_struct_hold = copy.copy(config_struct)
                
                coverage_struct,tile_struct = gwemopt.scheduler.schedule_alternating(params_hold, config_struct_hold, telescope, map_struct_hold, tile_struct,previous_coverage_struct)
                tile_struct, doReschedule,balanced_fields = gwemopt.utils.balance_tiles(params_hold, tile_struct, coverage_struct)
                config_struct_hold = copy.copy(config_struct)

                if doReschedule:
                    coverage_struct,tile_struct = gwemopt.scheduler.schedule_alternating(params_hold, config_struct_hold, telescope, map_struct_hold, tile_struct,previous_coverage_struct)
            
                # run a few more iterations to balance fields if max tiles is not optimized
                if not optimized_bool:
                    cnt,ntrials = 0,5
                    coverage_struct_hold = coverage_struct
                    while cnt < ntrials:
                        params_hold = copy.copy(params)
                        config_struct_hold = copy.copy(config_struct)
                        
                        tile_struct,doReschedule,balanced_fields = gwemopt.utils.balance_tiles(params_hold, tile_struct, coverage_struct_hold)
                        
                        if cnt == 0: before_balance = balanced_fields # get # of balanced fields before additional iters.
                        
                        if doReschedule:
                            coverage_struct_hold,tile_struct = gwemopt.scheduler.schedule_alternating(params_hold, config_struct_hold, telescope, map_struct_hold, tile_struct,previous_coverage_struct)
                            cnt += 1
                        else: break
                    
                    #use new coverage struct iff balanced fields have increased
                    if balanced_fields>before_balance:
                        coverage_struct = coverage_struct_hold


        else:

            #load the sun retriction for a satelite
            try:   
                sat_sun_restriction = config_struct["sat_sun_restriction"]
            except:
                sat_sun_restriction = 0.0
            print("sat_sun_restriction =", sat_sun_restriction)
            print("params[tilesType] =", params["tilesType"])
            if not params["tilesType"] == "galaxy":
                tile_struct = gwemopt.tiles.powerlaw_tiles_struct(params, config_struct, telescope, map_struct_hold, tile_struct)  
    
            elif sat_sun_restriction == 0.0:
                for key in tile_struct.keys():
                    # Check that a given tile is observable a minimum amount of time
                    # If not set the proba associated to the tile to zero
                    if 'segmentlist' and 'prob' in tile_struct[key] and tile_struct[key]['segmentlist'] and min_obs_duration > 0.0:
                        observability_duration = 0.0
                        for counter in range(len(tile_struct[key]['segmentlist'])):
                            observability_duration += tile_struct[key]['segmentlist'][counter][1] - tile_struct[key]['segmentlist'][counter][0]
                        if tile_struct[key]['prob'] > 0.0 and observability_duration < min_obs_duration:
                            tile_struct[key]['prob'] = 0.0
            
            else:
                # Check that a given tile is not to close to the sun for the satelite
                # If it's to close set the proba associated to the tile to zero

                time = map_struct["trigtime"]
                time = Time(time, format='isot', scale='utc')
                sun_position = get_sun(time)
                
                #astropy don't like the output of get sun in the following separator function, need here to redefine the skycoord
                sun_position = SkyCoord(sun_position.ra, sun_position.dec, frame='gcrs') 
       
                for key in tile_struct.keys():      
                    if 'segmentlist' and 'prob' in tile_struct[key] and tile_struct[key]['segmentlist']:

                        for counter in range(len(tile_struct[key]['segmentlist'])):

                            tile_position = SkyCoord(ra=tile_struct[key]['ra']*u.degree, dec=tile_struct[key]['dec']*u.degree, frame='icrs')
                            ang_dist = sun_position.separation(tile_position).deg

                        if ang_dist < sat_sun_restriction:

                            tile_struct[key]['prob'] = 0.0

          
            if params["timeallocationType"] == "manual": #only works if using same telescope
                try:
                    for field_id in params["observedTiles"]:
                        field_id = int(field_id)
                        if field_id in tile_struct:
                            tile_struct[field_id]['prob'] = 0.0
            
                except:
                    raise ValueError("need to specify tiles that have been observed using --observedTiles")
        
            if (params["doUpdateScheduler"] or params["doTreasureMap"]) and previous_coverage_struct:
                tile_struct = update_observed_tiles(params,tile_struct,previous_coverage_struct) #coverage_struct of the previous round

            if params["doSuperSched"]:
                tile_struct = erase_observed_tiles(params,tile_struct,telescope)

            coverage_struct = gwemopt.scheduler.scheduler(params, config_struct, tile_struct)

            if params["doBalanceExposure"]:
                cnt,ntrials = 0,10
                while cnt < ntrials:
                    tile_struct, doReschedule,balanced_fields = gwemopt.utils.balance_tiles(params, tile_struct, coverage_struct)
                    if doReschedule:
                        coverage_struct = gwemopt.scheduler.scheduler(params, config_struct, tile_struct)
                        cnt = cnt+1
                    else: break

            if params["doMaxTiles"]:
                tile_struct, doReschedule = gwemopt.utils.slice_number_tiles(params, telescope, tile_struct, coverage_struct)    
                if doReschedule:
                    coverage_struct = gwemopt.scheduler.scheduler(params, config_struct, tile_struct)

        tile_structs[telescope] = tile_struct
        coverage_structs.append(coverage_struct)

        if params["doIterativeTiling"]:
            map_struct_hold = gwemopt.utils.slice_map_tiles(params, map_struct_hold, coverage_struct)
               
    map_struct["prob"] = full_prob_map

    if params["doMovie_supersched"]:
        gwemopt.plotting.doMovie_supersched(params,combine_coverage_structs(coverage_structs),tile_structs,map_struct)
    
    return tile_structs, combine_coverage_structs(coverage_structs)

def pem(params, map_struct, tile_structs):

    map_struct_hold = copy.deepcopy(map_struct)

    coverage_structs = []
    for telescope in params["telescopes"]:
        config_struct = params["config"][telescope]
        tile_struct = tile_structs[telescope]

        if params["doAlternatingFilters"]:
            filters, exposuretimes = params["filters"], params["exposuretimes"]
            tile_struct_hold = copy.copy(tile_struct)
            coverage_structs_hold = []
            for filt, exposuretime in zip(filters,exposuretimes):
                params["filters"] = [filt]
                params["exposuretimes"] = [exposuretime]
                tile_struct_hold = gwemopt.tiles.pem_tiles_struct(params, config_struct, telescope, map_struct_hold, tile_struct_hold)
                coverage_struct_hold = gwemopt.scheduler.scheduler(params, config_struct, tile_struct_hold)
                coverage_structs_hold.append(coverage_struct_hold)
            coverage_struct = combine_coverage_structs(coverage_structs_hold)
        else:
            tile_struct = gwemopt.tiles.pem_tiles_struct(params, config_struct, telescope, map_struct_hold, tile_struct)
            coverage_struct = gwemopt.scheduler.scheduler(params, config_struct, tile_struct)
        coverage_structs.append(coverage_struct)

        if params["doIterativeTiling"]:
            map_struct_hold = gwemopt.utils.slice_map_tiles(map_struct_hold, coverage_struct)

    return combine_coverage_structs(coverage_structs)

def erase_observed_tiles(params,tile_struct,telescope): #only for run_gwemopt_superscheduler
    done_telescopes = []

    if len(params["coverage_structs"]) == 1: return tile_struct
    else:
        ii = len(params["coverage_structs"])-1 #finds out which round we are in

    while ii>0: #loops through all previous coverage structs + covered field ids to set tile probabilities to 0
        ii-=1
        prevtelescopes = params["alltelescopes"][ii].split(",")
        coverage_struct = params["coverage_structs"][f'coverage_struct_{ii}']
        if not coverage_struct: continue
        tile_struct_hold = gwemopt.utils.check_overlapping_tiles(params,tile_struct,coverage_struct)

        for prevtelescope in prevtelescopes:
            if prevtelescope in done_telescopes: continue #to prevent setting tiles to 0 redundantly
            done_telescopes.append(prevtelescope)

            if prevtelescope == telescope:
                for field_id in params["covered_field_ids"][prevtelescope][ii]:
                    tile_struct[field_id]['prob'] = 0.0
                continue

            for key in tile_struct.keys(): #maps field ids to tile_struct if not for the same telesocpe
                if not 'epochs' in tile_struct_hold[key]: continue
                epochs = tile_struct_hold[key]["epochs"]
                for jj in range(len(epochs)):
                    field_id = epochs[jj,5]
                    coverage_telescope = tile_struct[key]["epochs_telescope"][jj]
                    if field_id in params["covered_field_ids"][prevtelescope][ii] and coverage_telescope == prevtelescope: #makes sure field id obtained from check_overlapping_tiles is for the correct telescope
                        tile_struct[key]['prob'] = 0.0

                        break
    return tile_struct

def update_observed_tiles(params,tile_struct,previous_coverage_struct):
    
    if not params["doAlternatingFilters"]:
        tile_struct = gwemopt.utils.check_overlapping_tiles(params,tile_struct,previous_coverage_struct) #maps field ids to tile_struct

    for key in tile_struct.keys(): #sets tile to 0 if previously observed
        if 'epochs' not in tile_struct[key]: continue
        ipix = tile_struct[key]["ipix"]
        
        tot_overlap = sum(tile_struct[key]["epochs_overlap"]) #sums over list of overlapping ipix lengths

        if params["doAlternatingFilters"]:
            #only takes into account fields with same filters for total overlap
            for ii,filt in enumerate(tile_struct[key]["epochs_filters"]):
                if filt != params["filters"][0]:
                    tot_overlap -= tile_struct[key]["epochs_overlap"][ii]

        rat = tot_overlap/len(ipix)

        if rat > 0.3:
            tile_struct[key]['prob']=0.0

    return tile_struct

def timeallocation(params, map_struct, tile_structs,previous_coverage_struct=None):
    
    if len(params["telescopes"]) > 1 and params["doOrderByObservability"]:
        gwemopt.utils.order_by_observability(params,tile_structs)

    if params["timeallocationType"] == "powerlaw":
        print("Generating powerlaw schedule...")
        
        if params["doTreasureMap"]:
            if not params["treasuremap_token"]:
                print("Must provide Treasure Map API Token.")
                exit(0)
            treasuremap_coverage = gwemopt.utils.get_treasuremap_pointings(params)
        
            if previous_coverage_struct and treasuremap_coverage["data"]:
                previous_coverage_struct["data"] = np.append(previous_coverage_struct["data"],treasuremap_coverage["data"],axis=0)
                previous_coverage_struct["filters"] = previous_coverage_struct["filters"] + treasuremap_coverage["filters"]
                previous_coverage_struct["ipix"] = previous_coverage_struct["ipix"] + treasuremap_coverage["ipix"]
            elif treasuremap_coverage["data"]:
                previous_coverage_struct = treasuremap_coverage

        if (params["doUpdateScheduler"] or params["doTreasureMap"]) and not previous_coverage_struct:
                print("\nNo previous observations were ingested.\n")

        if params["doBlocks"]:
            exposurelists = {}
            scheduled_fields = {}
            for jj, telescope in enumerate(params["telescopes"]):
                config_struct = params["config"][telescope]
                exposurelist_split = np.array_split(config_struct["exposurelist"], params["Nblocks"])
                exposurelists[telescope] = exposurelist_split            
                scheduled_fields[telescope] = []
            tile_structs_hold = copy.copy(tile_structs)
            coverage_structs = []
            
            for ii in range(params["Nblocks"]):
                params_hold = copy.copy(params)
                params_hold["scheduled_fields"] = scheduled_fields
                for jj, telescope in enumerate(params["telescopes"]):
                    exposurelist = segments.segmentlist()
                    for seg in exposurelists[telescope][ii]:
                        exposurelist.append(segments.segment(seg[0],seg[1]))
                    params_hold["config"][telescope]["exposurelist"] = exposurelist
                    tile_structs_hold[telescope] = gwemopt.tiles.powerlaw_tiles_struct(params_hold, config_struct, telescope, map_struct, tile_structs_hold[telescope])
                tile_structs_hold, coverage_struct = gwemopt.coverage.powerlaw(params_hold, map_struct, tile_structs_hold,previous_coverage_struct)

                coverage_structs.append(coverage_struct)
                for ii in range(len(coverage_struct["ipix"])):
                    telescope = coverage_struct["telescope"][ii]
                    scheduled_fields[telescope].append(coverage_struct["data"][ii,5]) #appends all scheduled fields to appropriate list

            coverage_struct = combine_coverage_structs(coverage_structs)
        else:
            tile_structs, coverage_struct = gwemopt.coverage.powerlaw(params, map_struct, tile_structs,previous_coverage_struct)

    elif params["timeallocationType"] == "waw":
        if params["do3D"]:
            print("Generating WAW schedule...")
            coverage_struct = gwemopt.coverage.waw(params, map_struct, tile_structs)
        else:
            raise ValueError("Need to enable --do3D for waw")

    elif params["timeallocationType"] == "manual":
        print("Generating manual schedule...")
        tile_structs, coverage_struct = gwemopt.coverage.powerlaw(params, map_struct, tile_structs)

    elif params["timeallocationType"] == "pem":
        print("Generating PEM schedule...")
        coverage_struct = gwemopt.coverage.pem(params, map_struct, tile_structs)

    return tile_structs, coverage_struct 




import os, sys, optparse
import numpy as np
import healpy as hp
import json

from datetime import datetime
from astropy.time import Time

import matplotlib
#matplotlib.rc('text', usetex=True)
matplotlib.use('Agg')
matplotlib.rcParams.update({'font.size': 16})
matplotlib.rcParams['contour.negative_linestyle'] = 'solid'
import matplotlib.pyplot as plt

import ligo.gracedb
from ligo.gracedb.rest import GraceDb, HTTPError
from ligo.gracedb.rest import GraceDbBasic

url = 'https://gracedb.ligo.org/api/'
#url = 'https://gracedb-test.ligo.org/apibasic/'

def get_event(params):

    g = GraceDbBasic()
    eventString = params["event"]
    event = g.superevent(params["event"])
    preferred_event = g.event(event.json()["preferred_event"])
    jsonfile = os.path.join(params["outputDir"], 'data.json')
    with open(jsonfile, 'w') as outfile:
        json.dump(preferred_event.json(), outfile)    

    with open(jsonfile, 'r') as f:
        eventinfo = json.load(f)

    event_files = g.files(params["event"]).json()
    for filename in list(event_files):
        if "fits.gz" in filename:
            outfilename = os.path.join(params["outputDir"], filename)
            outfile = open(outfilename,'wb')
            r = g.files(params["event"], filename)
            outfile.write(r.read())
            outfile.close()
            skymapfile = outfilename

    return skymapfile, eventinfo
 




import os, sys
import numpy as np
import healpy as hp
from astropy.time import Time
import scipy.stats
from scipy.interpolate import interpolate as interp

import matplotlib
#matplotlib.rc('text', usetex=True)
matplotlib.use('Agg')
matplotlib.rcParams.update({'font.size': 16})
matplotlib.rcParams['contour.negative_linestyle'] = 'solid'
import matplotlib.pyplot as plt

import gwemopt.utils

def compute_efficiency(params, map_struct, lightcurve_struct, coverage_struct):

    nside = params["nside"]
    npix = hp.nside2npix(nside)

    Ninj = params["Ninj"]
    Ndet = params["Ndet"]
    gpstime = params["gpstime"]
    mjd_inj = Time(gpstime, format='gps', scale='utc').mjd
    #FOV_r = np.sqrt(float(params['FOV'])/np.pi)

    if params["doCatalog"]:
        distn = scipy.stats.rv_discrete(values=(np.arange(npix),
                                                map_struct["prob_catalog"]))
    else:
        distn = scipy.stats.rv_discrete(values=(np.arange(npix),
                                                map_struct["prob"]))
    ipix = distn.rvs(size=Ninj)
    ras, decs = hp.pix2ang(nside, ipix, lonlat=True)

    dists = np.logspace(-1,3,1000)
    ndetections = np.zeros((len(dists),))

    for pinpoint in ipix:
        idxs = []
        for jj in range(len(coverage_struct["ipix"])):
            expPixels = coverage_struct["ipix"][jj]
             
            if pinpoint in expPixels:
                idxs.append(jj)
        if len(idxs) == 0:
            continue

        mjds = coverage_struct["data"][idxs,2]
        mags = coverage_struct["data"][idxs,3]
        filts = coverage_struct["filters"][idxs]

        detections = np.zeros((len(dists),))

        for mjd, mag, filt in zip(mjds,mags,filts):
            lightcurve_t = lightcurve_struct["t"] + mjd_inj
            lightcurve_mag = lightcurve_struct[filt]
            idx = np.where(np.isfinite(lightcurve_mag))[0]

            f = interp.interp1d(lightcurve_t[idx], lightcurve_mag[idx],
                                fill_value='extrapolate')
            lightcurve_mag_interp = f(mjd)
            dist_threshold = (10**(((mag-lightcurve_mag_interp)/5.0)+1.0))/1e6

            idxs_detections = np.where(dists <= dist_threshold)[0]
            detections[idxs_detections] = detections[idxs_detections] + 1

        idxs_detections = np.where(detections >= Ndet)[0]
        ndetections[idxs_detections] = ndetections[idxs_detections] + 1

    efficiency = ndetections / Ninj
    efficiency_struct = {}
    efficiency_struct["ra"] = ras
    efficiency_struct["dec"] = decs
    efficiency_struct["efficiency"] = efficiency
    efficiency_struct["distances"] = dists

    save_efficiency_data(params, efficiency_struct, lightcurve_struct)
    efficiency_metric = calculate_efficiency_metric(params, efficiency_struct)
    save_efficiency_metric(params, os.path.join(params["outputDir"],"efficiency.txt"), efficiency_metric, lightcurve_struct)

    return efficiency_struct

def compute_3d_efficiency(params, map_struct, lightcurve_struct, coverage_struct):

    nside = params["nside"]
    npix = hp.nside2npix(nside)
    Ninj = params["Ninj"]
    gpstime = params["gpstime"]
    mjd_inj = Time(gpstime, format='gps', scale='utc').mjd
    
    if params["doCatalog"]:
        distn = scipy.stats.rv_discrete(values=(np.arange(npix),map_struct["prob_catalog"]))
    else:
        distn = scipy.stats.rv_discrete(values=(np.arange(npix),map_struct["prob"]))
    ipix = distn.rvs(size=Ninj)

    detections = 0

    for pinpoint in ipix:
        dist = -1
        while (dist < 0):
            dist = scipy.stats.norm(map_struct["distmu"][pinpoint],map_struct["distsigma"][pinpoint]).rvs()
        
        idxs = []
        for jj in range(len(coverage_struct["ipix"])):
            expPixels = coverage_struct["ipix"][jj]
            
            if pinpoint in expPixels:
                idxs.append(jj)

        if len(idxs) == 0:
            continue
    
        mjds = coverage_struct["data"][idxs,2]
        mags = coverage_struct["data"][idxs,3]
        filts = coverage_struct["filters"][idxs]
        
        for mjd, mag, filt in zip(mjds,mags,filts):
            lightcurve_t = lightcurve_struct["t"] + mjd_inj
            lightcurve_mag = lightcurve_struct[filt]
            idx = np.where(np.isfinite(lightcurve_mag))[0]
            
            f = interp.interp1d(lightcurve_t[idx], lightcurve_mag[idx],fill_value='extrapolate')
            lightcurve_mag_interp = f(mjd)
            dist_threshold = (10**(((mag-lightcurve_mag_interp)/5.0)+1.0))/1e6

            if dist<=dist_threshold:
                detections+=1

    print(f'Percent detections out of {Ninj} injected KNe: {detections*100/Ninj}% ')

def save_efficiency_data(params, efficiency_struct, lightcurve_struct):
    for i in range(0, len(efficiency_struct["distances"])):
        dist = efficiency_struct["distances"][i]
        eff = efficiency_struct["efficiency"][i]
        filename = os.path.join(params["outputDir"],'efficiency_' + lightcurve_struct['name'] + '.txt')
        if os.path.exists(filename):
            append_write = 'a'
            efficiency_file = open(filename, append_write)
        else:
            append_write = 'w'
            efficiency_file = open(filename, append_write)
            efficiency_file.write("Distance" + "\t" + "efficiency\n")
        efficiency_file.write(str(dist) + "\t" + str(eff) + "\n")
    efficiency_file.close()


def calculate_efficiency_metric(params, efficiency_struct):
    dist_sum = 0
    weighted_sum = 0
    for i in range(0, len(efficiency_struct["distances"])):
        dist = efficiency_struct["distances"][i]
        eff = efficiency_struct["efficiency"][i]
        dist_sum += dist * dist
        weighted_sum += dist * dist * eff
    metric = weighted_sum / dist_sum
    uncertainty = np.sqrt(metric * (1 - metric) / params["Ninj"])
    return (metric, uncertainty)


def save_efficiency_metric(params, efficiency_filename, efficiency_metric, lightcurve_struct):
    if os.path.exists(efficiency_filename):
        append_write = 'a'
        efficiency_file = open(efficiency_filename, append_write)
    else:
        append_write = 'w'
        efficiency_file = open(efficiency_filename, append_write)
        efficiency_file.write("tilesType\t" + "timeallocationType\t" + "scheduleType\t" + "Ntiles\t" + "efficiencyMetric\t" + "efficiencyMetric_err\t" + "injection\n")
    efficiency_file.write(params["tilesType"] + "\t" + params["timeallocationType"] + "\t" +  params["scheduleType"] + "\t" + str(params["Ntiles"]) + "\t" + str(efficiency_metric[0]) + "\t" + str(efficiency_metric[1]) + "\t" + lightcurve_struct["name"] + "\n")
    efficiency_file.close()



import datetime
import os
import glob
import copy

from astropy import table
from astropy import time
import ephem
import gwemopt.utils
import gwemopt.moc
import gwemopt.gracedb
import gwemopt.rankedTilesGenerator
import gwemopt.waw
import gwemopt.lightcurve
import gwemopt.coverage
import gwemopt.efficiency
import gwemopt.plotting
import gwemopt.tiles
import gwemopt.segments
import gwemopt.catalog
import healpy as hp
from ligo import segments
import numpy as np

def params_struct(skymap, gpstime, tobs=None, filt=['r'],
                  exposuretimes=[60.0],
                  mindiff=30.0*60.0, probability=0.9, tele='ZTF',
                  schedule_type='greedy',
                  doReferences=True,
                  filterScheduleType='block'):

    gwemoptpath = os.path.dirname(gwemopt.__file__)
    config_directory = os.path.join(gwemoptpath, '../config')
    tiling_directory = os.path.join(gwemoptpath, '../tiling')
    catalog_directory = os.path.join(gwemoptpath, '../catalog')

    params = {}
    params["config"] = {}
    config_files = glob.glob("%s/*.config" % config_directory)
    for config_file in config_files:
        telescope = config_file.split("/")[-1].replace(".config", "")
        if not telescope == tele: continue
        params["config"][telescope] =\
            gwemopt.utils.readParamsFromFile(config_file)
        params["config"][telescope]["telescope"] = telescope
        if "tesselationFile" in params["config"][telescope]:
            params["config"][telescope]["tesselationFile"] =\
                os.path.join(config_directory,
                             params["config"][telescope]["tesselationFile"])
            tesselation_file = params["config"][telescope]["tesselationFile"]
            if not os.path.isfile(tesselation_file):
                if params["config"][telescope]["FOV_type"] == "circle":
                    gwemopt.tiles.tesselation_spiral(
                        params["config"][telescope])
                elif params["config"][telescope]["FOV_type"] == "square":
                    gwemopt.tiles.tesselation_packing(
                        params["config"][telescope])

            params["config"][telescope]["tesselation"] =\
                np.loadtxt(params["config"][telescope]["tesselationFile"],
                           usecols=(0, 1, 2), comments='%')

        if "referenceFile" in params["config"][telescope]:
            params["config"][telescope]["referenceFile"] =\
                os.path.join(config_directory,
                             params["config"][telescope]["referenceFile"])
            refs = table.unique(table.Table.read(
                params["config"][telescope]["referenceFile"],
                format='ascii', data_start=2, data_end=-1)['field', 'fid'])
            reference_images =\
                {group[0]['field']: group['fid'].astype(int).tolist()
                 for group in refs.group_by('field').groups}
            reference_images_map = {1: 'g', 2: 'r', 3: 'i', 4: 'z', 5: 'J'}
            for key in reference_images:
                reference_images[key] = [reference_images_map.get(n, n)
                                         for n in reference_images[key]]
            params["config"][telescope]["reference_images"] = reference_images

        observer = ephem.Observer()
        observer.lat = str(params["config"][telescope]["latitude"])
        observer.lon = str(params["config"][telescope]["longitude"])
        observer.horizon = str(-12.0)
        observer.elevation = params["config"][telescope]["elevation"]
        params["config"][telescope]["observer"] = observer

    params["skymap"] = skymap
    params["gpstime"] = gpstime
    params["outputDir"] = "./" 
    params["tilingDir"] = tiling_directory
    params["event"] = ""
    params["telescopes"] = [tele]
    if tele in ["KPED", "GROWTH-India"]:
        params["tilesType"] = "galaxy"
        params["catalogDir"] = catalog_directory
        params["galaxy_catalog"] = "GLADE"
        params["galaxy_grade"] = "S"
        params["writeCatalog"] = False
        params["catalog_n"] = 1.0
        params["powerlaw_dist_exp"] = 1.0
    elif tele in ['TRE']:
        params["tilesType"] = "moc"
    elif tele in ['TNT']:
        params["tilesType"] = "galaxy"
        params["catalogDir"] = catalog_directory
        params["galaxy_catalog"] = "GLADE"
        params["galaxy_grade"] = "Sloc"
        params["writeCatalog"] = False
        params["catalog_n"] = 1.0
        params["powerlaw_dist_exp"] = 1.0
    else:
        params["tilesType"] = "moc"
    params["scheduleType"] = schedule_type
    params["timeallocationType"] = "powerlaw"
    params["nside"] = 256
    params["powerlaw_cl"] = probability
    params["powerlaw_n"] = 1.0
    params["powerlaw_dist_exp"] = 0.0

    params["galaxies_FoV_sep"] = 0.0

    params["doPlots"] = False
    params["doMovie"] = False
    params["doObservability"] = True
    params["do3D"] = False

    params["doFootprint"] = False
    params["footprint_ra"] = 30.0
    params["footprint_dec"] = 60.0
    params["footprint_radius"] = 10.0

    params["airmass"] = 2.5

    params["doCommitDatabase"] = True
    params["doRequestScheduler"] = False
    params["doEvent"] = False
    params["doSkymap"] = True
    params["doFootprint"] = False
    params["doDatabase"] = False
    params["doReferences"] = doReferences
    params["doChipGaps"] = False
    params["doSplit"] = False
    params["doParallel"] = False
    params["doUseCatalog"] = False
    params["doIterativeTiling"] = False
    params["doMinimalTiling"] = False
    params["doOverlappingScheduling"] = False

    params = gwemopt.utils.params_checker(params)

    if params["doEvent"]:
        params["skymap"], eventinfo = gwemopt.gracedb.get_event(params)
        params["gpstime"] = eventinfo["gpstime"]
        event_time = time.Time(params["gpstime"], format='gps', scale='utc')
        params["dateobs"] = event_time.iso
    elif params["doSkymap"]:
        event_time = time.Time(params["gpstime"], format='gps', scale='utc')
        params["dateobs"] = event_time.iso
    elif params["doFootprint"]:
        params["skymap"] = gwemopt.footprint.get_skymap(params)
        event_time = time.Time(params["gpstime"], format='gps', scale='utc')
        params["dateobs"] = event_time.iso
    elif params["doDatabase"]:
        event_time = time.Time(params["dateobs"], format='datetime',
                               scale='utc')
        params["gpstime"] = event_time.gps
    else:
        raise ValueError('Need to enable --doEvent, --doFootprint, '
                         '--doSkymap, or --doDatabase')

    if tobs is None:
        now_time = time.Time.now()
        timediff = now_time.gps - event_time.gps
        timediff_days = timediff / 86400.0
        params["Tobs"] = np.array([timediff_days, timediff_days+1])
    else:
        params["Tobs"] = tobs

    params["doSingleExposure"] = True
    if filterScheduleType == "block":
        params["doAlternatingFilters"] = True
    else:
        params["doAlternatingFilters"] = False
    params["filters"] = filt
    params["exposuretimes"] = exposuretimes
    params["mindiff"] = mindiff

    params = gwemopt.segments.get_telescope_segments(params)
    params['map_struct'] = None

    if params["doPlots"]:
        if not os.path.isdir(params["outputDir"]):
            os.makedirs(params["outputDir"])

    return params


def gen_structs(params):

    print('Loading skymap')
    # Function to read maps
    map_struct = gwemopt.utils.read_skymap(params, is3D=params["do3D"],
                                           map_struct=params['map_struct'])

    if params["tilesType"] == "galaxy":
        print("Generating catalog...")
        map_struct, catalog_struct =\
            gwemopt.catalog.get_catalog(params, map_struct)

    if params["tilesType"] == "moc":
        print('Generating MOC struct')
        moc_structs = gwemopt.moc.create_moc(params)
        tile_structs = gwemopt.tiles.moc(params, map_struct, moc_structs)
    elif params["tilesType"] == "ranked":
        print('Generating ranked struct')
        moc_structs = gwemopt.rankedTilesGenerator.create_ranked(params,
                                                                 map_struct)
        tile_structs = gwemopt.tiles.moc(params, map_struct, moc_structs)
    elif params["tilesType"] == "hierarchical":
        print('Generating hierarchical struct')
        tile_structs = gwemopt.tiles.hierarchical(params, map_struct)
    elif params["tilesType"] == "greedy":
        print('Generating greedy struct')
        tile_structs = gwemopt.tiles.greedy(params, map_struct)
    elif params["tilesType"] == "galaxy":
        print("Generating galaxy struct...")
        tile_structs = gwemopt.tiles.galaxy(params, map_struct, catalog_struct)
    else:
        raise ValueError(
            'Need tilesType to be moc, greedy, hierarchical, galaxy or ranked')

    coverage_struct = gwemopt.coverage.timeallocation(params,
                                                      map_struct,
                                                      tile_structs)

    if params["doPlots"]:
        gwemopt.plotting.skymap(params, map_struct)
        gwemopt.plotting.tiles(params, map_struct, tile_structs)
        gwemopt.plotting.coverage(params, map_struct, coverage_struct)

    return map_struct, tile_structs, coverage_struct


def test_scheduler():
    # Read test GCN

    gwemoptpath = os.path.dirname(gwemopt.__file__)
    testpath = config_directory = os.path.join(gwemoptpath, 'tests')
    skymap = os.path.join(testpath, 'data/MS190227n_bayestar.fits.gz')
    gpstime = 1235311089.400738

    teles=['ZTF','KPED', 'TRE', 'TNT']
    doReferences_list = [True,False, False, False]
    for tele,doReferences in zip(teles,doReferences_list): 
        params = params_struct(skymap, gpstime, tele=tele,
                               doReferences=doReferences)
        map_struct, tile_structs, coverage_struct = gen_structs(params)
